# $Id: test_lib_utils.py 68b0cb3c67d5 2011/12/18 18:50:45 patrick $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``lib.utils`` classes and functions."""

import unittest
from os.path import join, dirname

INI_FILE = join(
    dirname(__file__), '..', 'Processors', 'Publidoc2Html', 'leprisme.ini')
CONTENT_DIR = join(dirname(__file__), '..', 'Processors', 'PublidocValid')
CONTENT_EXCLUDE = ('.hg', 'publiset.rng')


# =============================================================================
class UnitTestLibUtilsConfigGet(unittest.TestCase):
    """Unit test class for ``lib.utils.config_get``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.config_get]"""
        from ConfigParser import ConfigParser
        from ..lib.utils import config_get
        config = ConfigParser({'here': dirname(INI_FILE)})
        config.read(INI_FILE)
        self.assertEqual(config_get(config, 'Input', 'file_regex'), r'\.xml$')
        self.assertEqual(config_get(config, 'Input', 'foo', 'bar'), 'bar')


# =============================================================================
class UnitTestLibUtilsCopyContent(unittest.TestCase):
    """Unit test class for ``lib.utils.copy_content``."""
    # pylint: disable = I0011, C0103

    # -------------------------------------------------------------------------
    def __init__(self, method_name='runTest'):
        """Constructor method."""
        super(UnitTestLibUtilsCopyContent, self).__init__(method_name)
        self._tmp_dir = None

    # -------------------------------------------------------------------------
    def setUp(self):
        """Create a temporary directory."""
        from tempfile import mkdtemp
        self.tearDown()
        self._tmp_dir = mkdtemp()

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from os.path import exists
        from shutil import rmtree
        if self._tmp_dir is not None and exists(self._tmp_dir):
            rmtree(self._tmp_dir)
        self._tmp_dir = None

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.copy_content]"""
        from os import walk
        from os.path import exists, relpath
        from ..lib.utils import copy_content
        copy_content(CONTENT_DIR, self._tmp_dir, CONTENT_EXCLUDE)
        for path, dirs, files in walk(CONTENT_DIR):
            for name in dirs + files:
                copy = join(self._tmp_dir, relpath(path, CONTENT_DIR), name)
                if name in CONTENT_EXCLUDE:
                    self.assertFalse(exists(copy))
                else:
                    self.assertTrue(exists(copy))


# =============================================================================
class UnitTestLibUtilsCamelCase(unittest.TestCase):
    """Unit test class for ``lib.utils.camel_case``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.camel_case]"""
        from ..lib.utils import camel_case
        self.assertEqual(camel_case('pdoc2html'), 'Pdoc2Html')
        self.assertEqual(camel_case('LaTeX'), 'LaTeX')
        self.assertEqual(camel_case('laTeX'), 'LaTeX')
        self.assertEqual(camel_case('my_way'), 'MyWay')
        self.assertEqual(camel_case('my way'), 'MyWay')
        self.assertEqual(camel_case('my-way'), 'My-Way')


# =============================================================================
class UnitTestLibUtilsHashSha(unittest.TestCase):
    """Unit test class for ``lib.utils.hash_sha``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.hash_sha]"""
        from ..lib.utils import hash_sha
        self.assertEqual(hash_sha('protectme', None),
                         '360a06e80743072020f69f2129c4933ea29f879f')
        self.assertEqual(hash_sha('protectme', 'seekrit'),
                         'e50267004c95597d525f9a16e68d046c80eb0ded')
        self.assertEqual(hash_sha(u'protègemoi', None),
                         'c4eb986ca838b6e2dc6c4b1bdfa31142ea565a84')


# =============================================================================
class UnitTestLibUtilsEncrypt(unittest.TestCase):
    """Unit test class for ``lib.utils.encrypt``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.encrypt]"""
        from ..lib.utils import encrypt
        self.assertEqual(encrypt('protectme', None),
                         'XzgU2PN974c3yYOp4FsJzw==')
        self.assertEqual(encrypt('protectme', 'seekrit'),
                         'B/48LdYirNqHOFq3dMIWQA==')
        self.assertEqual(encrypt('protègemoi', 'seekrit'),
                         'GoJ3XiMjxaXqm6eDr9anfg==')


# =============================================================================
class UnitTestLibUtilsDecrypt(unittest.TestCase):
    """Unit test class for ``lib.utils.decrypt``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.decrypt]"""
        from ..lib.utils import decrypt
        self.assertEqual(decrypt('XzgU2PN974c3yYOp4FsJzw==', None),
                         'protectme')
        self.assertEqual(decrypt('B/48LdYirNqHOFq3dMIWQA==', 'seekrit'),
                         'protectme')
        self.assertEqual(decrypt('GoJ3XiMjxaXqm6eDr9anfg==', 'seekrit'),
                         'protègemoi')



# $Id: test_models_users.py df43b426606b 2012/03/09 15:58:44 yvon $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``models.users`` classes."""

import unittest

from pyramid import testing

from ..tests import ModelTestCase


# =============================================================================
class UnitTestModelsUsersUser(ModelTestCase):
    """Unit test class for ``models.users.User``."""
    # pylint: disable = I0011, C0103, R0904

    _key = 'seekrit'
    _password = 'mypassword'

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.users import User
        self.session.query(User).filter(
            User.login.like('_test_%')).delete('fetch')
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make an ``User`` object."""
        from ..models.users import User
        return User({'auth.key': self._key}, '_test_user', status='active',
                    password=self._password, name=u'Félicie TASSION',
                    email='test@prismallia.fr', lang='en')

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.users.User.__init__]"""
        from ..models.users import User
        from datetime import datetime
        from ..lib.utils import hash_sha
        user1 = self._make_one()
        self.session.add(user1)
        self.session.commit()
        user2 = self.session.query(User).filter_by(login=user1.login).first()
        self.assertNotEqual(user2, None)
        self.assertEqual(user2.login, user1.login)
        self.assertEqual(user2.status, user1.status)
        self.assertEqual(user2.password, hash_sha(self._password, self._key))
        self.assertEqual(user2.name, user1.name)
        self.assertEqual(user2.email, user1.email)
        self.assertEqual(user2.lang, user1.lang)
        self.assertTrue(user2.created <= datetime.now())
        self.assertTrue(user2.created == user2.updated)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.users.User.__repr__]"""
        user = self._make_one()
        info = repr(user)
        self.assertTrue('<User(' in info)
        self.assertTrue('-1' in info)
        self.assertTrue('%s' % user.login in info)
        self.assertTrue('%s' % user.status in info)
        self.assertTrue('%s' % user.name.encode('utf8') in info)
        self.assertTrue('%s' % user.lang in info)

    # -------------------------------------------------------------------------
    def test_set_password(self):
        """[u:models.users.User.set_password]"""
        from ..lib.utils import hash_sha
        user = self._make_one()
        password = user.password
        user.set_password({'auth.key': self._key}, 'foo')
        self.assertNotEqual(user.password, password)
        self.assertEqual(user.password, hash_sha('foo', self._key))

    # -------------------------------------------------------------------------
    def test_setup_environment(self):
        """[u:models.users.User.setup_environment]"""
        request = testing.DummyRequest()
        request.registry.settings['auth.key'] = self._key
        request.registry.settings['available_languages'] = 'fr en'
        user = self._make_one()
        user.setup_environment(request)
        self.assertTrue('user_id' in request.session)
        self.assertEqual(request.session['user_id'], user.user_id)
        self.assertTrue('name' in request.session)
        self.assertEqual(request.session['name'], user.name)
        self.assertTrue('paging' in request.session)
        self.assertEqual(request.session['paging'][0], user.page_size)
        self.assertTrue('home' in request.session)
        self.assertEqual(request.session['home'], user.home)


# =============================================================================
class UnitTestModelsUsersUserPerm(unittest.TestCase):
    """Unit test class for ``models.users.UserPerm``."""

    _user_id = 1000

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``UserPerm`` object."""
        from ..models.users import UserPerm
        return UserPerm(self._user_id, 'usr', 'manager')

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.users.UserPerm.__init__]"""
        user_perm = self._make_one()
        self.assertEqual(user_perm.scope, 'usr')
        self.assertEqual(user_perm.level, 'manager')

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.users.UserPerm.__repr__]"""
        user_perm = self._make_one()
        info = repr(user_perm)
        self.assertTrue('<UserPerm(' in info)
        self.assertTrue(str(self._user_id) in info)
        self.assertTrue("'%s'" % user_perm.scope in info)
        self.assertTrue("'%s'" % user_perm.level in info)


# =============================================================================
class UnitTestModelsUsersUserIP(unittest.TestCase):
    """Unit test class for ``models.users.UserIP``."""

    _user_id = 1000

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make an ``UserIP`` object."""
        from ..models.users import UserIP
        return UserIP(self._user_id, '192.168.0.1')

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.users.UserIP.__init__]"""
        user_ip = self._make_one()
        self.assertEqual(user_ip.ip, '192.168.0.1')

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.users.UserIP.__repr__]"""
        user_ip = self._make_one()
        info = repr(user_ip)
        self.assertTrue('<UserIP(' in info)
        self.assertTrue(str(self._user_id) in info)
        self.assertTrue('%s' % user_ip.ip in info)


# =============================================================================
class IntegrationTestModelsUsers(ModelTestCase):
    """Integration test class for ``models.users``."""
    # pylint: disable = I0011, C0103

    _user_id = 1000

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.users import User
        self.session.query(User).filter(
            User.login.like('_test_%')).delete('fetch')
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_user(self):
        """Make an ``User`` object."""
        # pylint: disable = I0011, E1101
        from ..models.users import User
        user = User({'auth.key': 'seekrit'}, '_test_user', status='active',
                    password='mypassword', name=u'Gérard MENVUSSA',
                    email='test@prismallia.fr', lang='en')
        user.user_id = self._user_id
        return user

    # -------------------------------------------------------------------------
    def test_delete_cascade_ip(self):
        """[i:models.users.User] delete, cascade IP"""
        # pylint: disable = I0011, W0104
        from ..models.users import User, UserIP
        user = self._make_user()
        user.ips.append(UserIP(user.user_id, '192.168.0.1'))
        self.session.add(user)
        self.session.commit()
        user_id = user.user_id
        user_ip = self.session.query(UserIP).filter_by(user_id=user_id).first()
        self.assertNotEqual(user_ip, None)
        self.assertEqual(user_ip.ip, '192.168.0.1')
        self.session.query(User).filter_by(user_id=user_id).delete()
        self.session.commit()
        user_ip = self.session.query(UserIP).filter_by(user_id=user_id).first()
        self.assertEqual(user_ip, None)



# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``models.processors`` classes."""

from os.path import dirname, join
from ..tests import ModelTestCase


# =============================================================================
class UnitTestModelsProcessorssProcessor(ModelTestCase):
    """Unit test class for ``models.processor.Processor``."""
    # pylint: disable = I0011, C0103

    _processor_path = join(dirname(__file__), 'Data', 'processor.xml')
    _processor_id = u'TestProcessor'
    _processor_xml = ' '
    _settings = {'storage.root': '/', 'agent.refresh': 180, \
                 'agent.0.weight': 1, 'agent.0.password': 'yo', \
                 'uid': 'devfront', 'pyramid.default_locale_name': 'en'}

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.processors import Processor
        self.session.query(Processor)\
            .filter_by(processor_id=self._processor_id).delete()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``Processor`` object."""
        from ..models.processors import Processor
        processor = Processor(self._processor_id)
        return processor

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.processors.Processor.__init__]"""
        from ..models.processors import Processor
        processor1 = self._make_one()
        with open(self._processor_path, 'r') as f:
            self._processor_xml = f.read()
        Processor.load(processor1.processor_id, self._processor_xml)
        self.session.add(processor1)
        self.session.commit()
        processor2 = self.session.query(Processor)\
                     .filter_by(processor_id=processor1.processor_id).first()
        self.assertNotEqual(processor2, None)
        self.assertEqual(processor2.processor_id, processor1.processor_id)
        self.assertEqual(processor2.xml, processor1.xml)
        self.assertEqual(processor2.description, processor1.description)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.processors.Processor.__repr__]"""
        processor = self._make_one()
        info = repr(processor)
        self.assertTrue('<Processor(' in info)
        self.assertTrue('%s' % self._processor_id in info)

    # -------------------------------------------------------------------------
    def test_labels(self):
        """[u:models.processors.Processor.labels]"""
        from ..models.processors import Processor

        processor1 = self._make_one()
        self.session.add(processor1)
        self.session.commit()
        with open(self._processor_path, 'r') as f:
            self._processor_xml = f.read()
        Processor.load(processor1.processor_id, self._processor_xml)
        processor2 = self.session.query(Processor)\
                     .filter_by(processor_id=processor1.processor_id).first()
        request = self._make_request()
        request.registry.settings = self._settings
        request.registry['fbuild'] = FrontBuildManager()
        labels_dic = Processor.labels(request)
        self.assertTrue(labels_dic[processor2.processor_id] is not None)
        self.assertEqual(processor2.processor_id, processor1.processor_id)

    # -------------------------------------------------------------------------
    def test_description(self):
        """[u:models.processors.Processor.description]"""
        from ..models.processors import Processor
        from lxml import etree
        from ..lib.xml import local_text
        from cStringIO import StringIO

        processor1 = self._make_one()
        self.session.add(processor1)
        self.session.commit()
        with open(self._processor_path, 'r') as f:
            self._processor_xml = f.read()
        Processor.load(processor1.processor_id, self._processor_xml)
        processor2 = self.session.query(Processor)\
                     .filter_by(processor_id=processor1.processor_id).first()
        request = self._make_request()
        request.session['lang'] = 'en'
        request.registry.settings = self._settings
        request.registry['fbuild'] = FrontBuildManager()
        description2 = Processor.description(request, processor2.processor_id)

        xml = etree.parse(StringIO(self._processor_xml))
        description1 = local_text(xml.find('processor'), \
                                  'description', request)
        self.assertTrue(description2 is not None)
        self.assertEqual(description1, description2)


# =============================================================================
class FrontBuildManager(object):
    """This class manages front builds.

    Fake FrontBuildManager for Testing.
    """
    # pylint: disable = I0011, R0902

    # -------------------------------------------------------------------------
    def __init__(self):
        """Constructor method.
        """

    # -------------------------------------------------------------------------
    @classmethod
    def refresh_agent_list(cls, request):
        """empty function to bypass agents network request.
        """
        # pylint: disable = I0011, W0613
        return



# $Id: publidoc2html_pre.py 1292397d8520 2011/12/30 17:33:03 patrick $
"""LePrisme preprocess script."""

from publiforge.lib.utils import _


# =============================================================================
def main(engine, filename, fid, data, output):
    """Preprocess script.

    :param engine: (:class:`Engine` instance)
        Engine object on which it depends.
    :param filename: (string)
        Relative path to the original file to transform.
    :param fid: (string)
        File identifier.
    :param data: (string or `:class:lxml.etree.ElementTree` object)
        Data to transform in a string or in a tree.
    :param output: (string)
        Full path to output directory.
    :return: (string, `:class:lxml.etree.ElementTree` object or ``None``)
        Modified data or ``None`` if fails.
    """
    # pylint: disable = I0011, W0613
    engine.build.log(_('${f}: dummy preprocess of "${f}"', {'f': fid}))
    return data



# $Id: project.py ada86c3ea608 2012/03/22 22:52:47 patrick $
# pylint: disable = I0011, C0322
"""Project view callables."""

from os.path import join, dirname, basename, commonprefix, relpath
from colander import Mapping, SchemaNode, Length, String, OneOf, Date
import zipfile
import tempfile
from sqlalchemy import select, desc, or_, and_

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPForbidden

from ..lib.utils import _, has_permission
from ..lib.xml import load, upload_configuration, export_configuration
from ..lib.views import get_action, paging_users, paging_groups
from ..lib.views import current_storage, current_project, file_download
from ..lib.form import Form
from ..lib.renderer import Paging, TabSet
from ..models import LABEL_LEN, DESCRIPTION_LEN, DBSession, close_dbsession
from ..models.users import User
from ..models.groups import GROUPS_USERS, Group
from ..models.projects import PROJECT_STATUS, PROJECT_PERMS, Project
from ..models.projects import ProjectUser, ProjectGroup
from ..models.projects import ProjectProcessing, ProjectPack


PROJECT_SETTINGS_TABS = (
    _('Description'), _('Processings'), _('Members'), _('Member groups'))


# =============================================================================
class ProjectView(object):
    """Class to manage projects."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='project_admin',
                 renderer='../Templates/prj_admin.pt',
                 permission='prj.update')
    def admin(self):
        """List projects for administration purpose."""
        action, items = get_action(self._request)
        if action == 'imp!':
            upload_configuration(self._request, 'prj_manager', 'project')
            action = ''
        elif action[0:4] == 'del!':
            self._delete_projects(items)
            action = ''
        elif action[0:4] == 'exp!':
            return self._export_projects(items)

        paging, defaults = self._paging_projects()
        form = Form(self._request, defaults=defaults)
        groups = [(k.group_id, k.label) for k in
                  DBSession.query(Group).join(ProjectGroup)]

        depth = (self._request.breadcrumbs.current_path() ==
                 self._request.route_path('site_admin') and 3) or 2
        self._request.breadcrumbs.add(_('Project administration'), depth)
        return {
            'form': form, 'paging': paging, 'action': action,
            'groups': groups, 'project_status': PROJECT_STATUS,
            'i_editor':  has_permission(self._request, 'prj_editor'),
            'i_manager': has_permission(self._request, 'prj_manager')}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_index',
                 renderer='../Templates/prj_index.pt',
                 permission='prj.read')
    def index(self):
        """List authorized projects."""
        action, items = get_action(self._request)
        if action[0:4] == 'exp!':
            return self._export_projects(items)
        elif action[0:3] == 'men':
            user_id = self._request.session['user_id']
            project_user = DBSession.query(ProjectUser).filter_by(
                project_id=action[4:], user_id=user_id).first()
            if project_user is None:
                project_user = ProjectUser(action[4:], user_id)
                DBSession.add(project_user)
            project_user.in_menu = action[3] == '+'
            DBSession.commit()
            del self._request.session['menu']

        paging, defaults = self._paging_projects(
            self._request.session['user_id'])
        form = Form(self._request, defaults=defaults)
        in_menus = [k[0] for k in DBSession.query(ProjectUser.project_id)
                    .filter_by(user_id=self._request.session['user_id'])
                    .filter_by(in_menu=True).all()]

        if 'container' in self._request.session:
            del self._request.session['container']

        self._request.breadcrumbs.add(_('All projects'), 2)
        return {
            'form': form, 'paging': paging, 'action': action,
            'project_status': PROJECT_STATUS, 'in_menus': in_menus}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_view',
                 renderer='../Templates/prj_view.pt',
                 permission='prj.read')
    def view(self):
        """Show project settings with its users."""
        action = get_action(self._request)[0]
        if action == 'exp!':
            return self._export_projects((
                self._request.matchdict.get('project_id'),))

        project = current_project(self._request, only_dict=False)
        tab_set = TabSet(self._request, PROJECT_SETTINGS_TABS)
        members = DBSession.query(
            User.user_id, User.login, User.name, ProjectUser.perm)\
            .join(ProjectUser)\
            .filter(ProjectUser.project_id == project.project_id)\
            .filter(ProjectUser.perm != None).order_by(User.login)
        member_groups = DBSession.query(
            Group.group_id, Group.label, ProjectGroup.perm)\
            .filter(ProjectGroup.project_id == project.project_id)\
            .filter(Group.group_id == ProjectGroup.group_id)\
            .order_by(Group.label).all()
        i_editor = has_permission(self._request, 'prj_editor') \
            or self._request.session['project']['perm'] == 'editor'

        self._request.breadcrumbs.add(
            _('Project settings'), replace=self._request.route_path(
            'project_edit', project_id=project.project_id))
        return {
            'form': Form(self._request), 'tab_set': tab_set,
            'PROJECT_STATUS': PROJECT_STATUS, 'perms': PROJECT_PERMS,
            'project': project, 'members': members,
            'member_groups': member_groups, 'i_editor': i_editor}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_create',
                 renderer='../Templates/prj_edit.pt',
                 permission='prj.create')
    def create(self):
        """Create a project."""
        form, tab_set = self._settings_form()

        if form.validate():
            label = ' '.join(form.values['label'].split()).strip()
            project = DBSession.query(Project).filter_by(label=label).first()
            if project is None:
                project = Project(label, form.values['description'],
                    form.values['status'], form.values['deadline'])
                DBSession.add(project)
                DBSession.commit()
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path(
                    'project_edit', project_id=project.project_id))
            self._request.session.flash(
                _('This label already exists.'), 'alert')

        self._request.breadcrumbs.add(_('Project creation'))
        return{
            'form': form, 'tab_set': tab_set,
            'PROJECT_STATUS': PROJECT_STATUS, 'perms': PROJECT_PERMS,
            'project': None, 'members': None, 'group_members': None,
            'paging': None, 'groups': None}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_edit',
                 renderer='../Templates/prj_edit.pt')
    def edit(self):
        """Edit project settings."""
        # Authorization
        project = current_project(self._request, only_dict=False)
        if not has_permission(self._request, 'prj_editor') \
               and self._request.session['project']['perm'] != 'editor':
            raise HTTPForbidden()

        # Action
        action = get_action(self._request)[0]
        paging = groups = None
        if action == 'aur?' or \
               (not action and self._request.GET.get('paging_id') == 'users'):
            paging = paging_users(self._request)[0]
            groups = DBSession.query(Group.group_id, Group.label).all()
        elif action[0:4] == 'aur!':
            self._add_members(project)
        elif action[0:4] == 'rur!':
            del self._request.session['menu']
            DBSession.query(ProjectUser).filter_by(
                project_id=project.project_id, user_id=int(action[4:]))\
                .delete()
            DBSession.commit()
        elif action == 'agp?' or (not action and
                self._request.GET.get('paging_id') == 'groups'):
            paging = paging_groups(self._request)[0]
        elif action[0:4] == 'agp!':
            self._add_member_groups(project)
        elif action[0:4] == 'rgp!':
            del self._request.session['menu']
            DBSession.query(ProjectUser).filter_by(
                project_id=project.project_id, user_id=int(action[4:]),
                perm=None).delete()
            DBSession.query(ProjectGroup).filter_by(
                project_id=project.project_id, group_id=action[4:]).delete()
            DBSession.commit()
        elif action == 'imp!':
            self._upload_processing(project.project_id)
        elif action[0:4] == 'del!':
            DBSession.query(ProjectProcessing).filter_by(
                project_id=project.project_id, processing_id=int(action[4:]))\
                .delete()
            DBSession.commit()
            if  'container' in self._request.session:
                del self._request.session['container']

        # Environment
        form, tab_set = self._settings_form(project)
        members = DBSession.query(
            User.user_id, User.login, User.name, ProjectUser.perm)\
            .filter(ProjectUser.project_id == project.project_id)\
            .filter(User.user_id == ProjectUser.user_id)\
            .order_by(User.login).all()
        member_groups = DBSession.query(
            Group.group_id, Group.label, ProjectGroup.perm)\
            .filter(ProjectGroup.project_id == project.project_id)\
            .filter(Group.group_id == ProjectGroup.group_id)\
            .order_by(Group.label).all()

        # Save
        view_path = self._request.route_path('project_view',
            project_id=project.project_id)
        if action == 'sav!' and form.validate(project) \
               and self._save(project, form.values):
            return HTTPFound(view_path)
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs.add(_('Project settings'), replace=view_path)

        return {
            'form': form, 'action': action, 'tab_set': tab_set,
            'PROJECT_STATUS': PROJECT_STATUS, 'perms': PROJECT_PERMS,
            'project': project, 'members': members,
            'member_groups': member_groups, 'paging': paging, 'groups': groups}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_dashboard',
                 renderer='../Templates/prj_dashboard.pt',
                 permission='prj.read')
    def dashboard(self):
        """Display a project dashboard."""
        project = current_project(self._request)
        action, items = get_action(self._request)
        if action[0:4] == 'bld!':
            return HTTPFound(self._request.route_path('build_launch',
                project_id=self._request.matchdict.get('project_id'),
                processing_id=self._request.params['processing_id'],
                pack_ids='_'.join(items)))
        elif action[0:4] == 'del!':
            DBSession.query(ProjectPack)\
                .filter_by(project_id=project['project_id'])\
                .filter(ProjectPack.pack_id.in_(items)).delete('fetch')
            DBSession.commit()
            action = ''

        paging, defaults = self._paging_packs(project['project_id'])
        defaults['processing_id'] = project['processing_id']
        form = Form(self._request, defaults=defaults)

        self._request.breadcrumbs.add(_('Dashboard'), 2)
        return {
            'form': form, 'paging': paging, 'action': action,
            'project': project}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_results',
                 renderer='../Templates/prj_results.pt',
                 permission='prj.read')
    def results(self):
        """List all builds in progress or with result."""
        # Build parameters
        project = current_project(self._request)
        build_list = self._request.registry['fbuild']\
                     .build_list(project['project_id'])
        working = prgrss = users = processings = packs = status_label = None
        if build_list:
            users = [k['user_id'] for k in build_list]
            users = dict(DBSession.query(User.user_id, User.name)
                         .filter(User.user_id.in_(users)).all())
            processings = [k['processing_id'] for k in build_list]
            processings = DBSession.query(ProjectProcessing)\
              .filter_by(project_id=project['project_id'])\
              .filter(ProjectProcessing.processing_id.in_(processings)).all()
            processings = dict([
                (k.processing_id, (k.label, k.output)) for k in processings])
            packs = [k['pack_id'] for k in build_list]
            packs = dict(DBSession.query(
                ProjectPack.pack_id, ProjectPack.label)
                         .filter_by(project_id=project['project_id'])
                         .filter(ProjectPack.pack_id.in_(packs)).all())
            status_label = {'none':  _('In progress...'), 'stop': _('Stopped'),
                            'end': _('Completed'), 'fatal': _('In error')}
            working, prgrss = self._request.registry['fbuild'].progress(
                self._request, [k['build_id'] for k in build_list])

        # Action
        form = Form(self._request)
        action, items = get_action(self._request)
        if action[0:4] == 'dnl!':
            action = self._download_results(build_list, processings, items)
            if action is not None:
                return action

        # Refresh
        if working and not 'ajax' in self._request.params:
            self._request.response.headerlist.append(('Refresh',
                self._request.registry.settings.get('refresh.long', '8')))

        self._request.breadcrumbs.add(_('Results'))
        return {
            'form': form, 'working': working, 'project': project,
            'results': build_list, 'progress': prgrss, 'users': users,
            'processings': processings, 'packs': packs,
            'status_label': status_label}

    # -------------------------------------------------------------------------
    @view_config(route_name='project_results_ajax', renderer='json',
                 permission='prj.read')
    def results_ajax(self):
        """List all builds in progress for AJAX request."""
        project_id = int(self._request.matchdict['project_id'])
        build_list = self._request.registry['fbuild'].build_list(project_id)
        working, prgrss = self._request.registry['fbuild'].progress(
            self._request, [k['build_id'] for k in build_list])
        response = dict([(k, prgrss[k][1]) for k in prgrss])
        response['working'] = working
        return response

    # -------------------------------------------------------------------------
    def _paging_projects(self, user_id=None):
        """Return a :class:`~..lib.renderer.Paging` object filled with
        projects.

        :param user_id: (integer, optional)
            Select only projects of user ``user_id``.
        :return: (tuple)
            A tuple such as ``(paging, filters)`` where ``paging`` is a
            :class:`~..lib.renderer.Paging` object and ``filters`` a
            dictionary of filters.
        """
        # Parameters
        paging_id = user_id is None and 'projects!' or 'projects'
        params = Paging.params(self._request, paging_id, '+label')
        if len(self._request.POST) == 0 and not 'f_status' in params:
            params['f_status'] = 'active'

        # Query
        query = DBSession.query(Project)
        if user_id is not None:
            groups = [k.group_id for k in DBSession.execute(
                select([GROUPS_USERS], user_id == GROUPS_USERS.c.user_id))]
            if groups:
                query = query.filter(or_(
                    and_(ProjectUser.user_id == user_id,
                         ProjectUser.perm != None,
                         ProjectUser.project_id == Project.project_id),
                    and_(ProjectGroup.group_id.in_(groups),
                         ProjectGroup.project_id == Project.project_id))
                    ).distinct(Project.project_id, Project.label)
            else:
                query = query.join(ProjectUser)\
                       .filter(ProjectUser.user_id == user_id)\
                       .filter(ProjectUser.perm != None)
        elif 'f_login' in params:
            query = query.join(ProjectUser).join(User)\
                    .filter(User.login == params['f_login'])\
                    .filter(ProjectUser.perm != None)
        if 'f_group' in params:
            query = query.join(ProjectGroup).filter(
                ProjectGroup.group_id == params['f_group'])
        if 'f_label' in params:
            query = query.filter(
                Project.label.ilike('%%%s%%' % params['f_label']))
        if 'f_status' in params and params['f_status'] != '*':
            query = query.filter(Project.status == params['f_status'])

        # Order by
        oby = 'projects.%s' % params['sort'][1:]
        query = query.order_by(desc(oby) if params['sort'][0] == '-' else oby)

        return Paging(self._request, paging_id, query), params

    # -------------------------------------------------------------------------
    def _paging_packs(self, project_id):
        """Return a :class:`~..lib.renderer.Paging` object filled with packs
        of project ``project_id``.

        :param project_id: (integer)
            Project ID.
        :return: (tuple)
            A tuple such as ``(paging, filters)`` where ``paging`` is a
            :class:`~..lib.renderer.Paging` object and ``filters`` a
            dictionary of filters.
        """
        # Parameters
        params = Paging.params(self._request, 'packs', '+pack_id')

        # Query
        query = DBSession.query(ProjectPack).filter_by(project_id=project_id)
        if 'f_label' in params:
            query = query.filter(
                ProjectPack.label.ilike('%%%s%%' % params['f_label']))

        # Order by
        oby = 'projects_packs.%s' % params['sort'][1:]
        query = query.order_by(desc(oby) if params['sort'][0] == '-' else oby)

        return Paging(self._request, 'packs', query), params

    # -------------------------------------------------------------------------
    def _delete_projects(self, project_ids):
        """Delete projects.

        :param project_ids: (list)
            List of project IDs to delete.
        """
        if not has_permission(self._request, 'prj_manager'):
            raise HTTPForbidden()
        DBSession.query(Project).filter(
            Project.project_id.in_(project_ids)).delete('fetch')
        DBSession.commit()
        del self._request.session['menu']

    # -------------------------------------------------------------------------
    def _export_projects(self, project_ids):
        """Export projects.

        :param project_ids: (list)
            List of project IDs to export.
        :return: (:class:`pyramid.response.Response` instance)
        """
        i_editor = has_permission(self._request, 'prj_editor')
        user_id = self._request.session['user_id']
        groups = set([k.group_id for k in DBSession.execute(
            select([GROUPS_USERS], GROUPS_USERS.c.user_id == user_id))])
        elements = []
        for project in DBSession.query(Project)\
                .filter(Project.project_id.in_(project_ids))\
                .order_by('label'):
            if i_editor or user_id in [k.user_id for k in project.users] \
                   or (groups and
                       groups & set([k.group_id for k in project.groups])):
                elements.append(project.xml(self._request))

        name = '%s_projects.pfp' % self._request.registry.settings.get(
            'skin.label', 'publiforge')
        return export_configuration(elements, name)

    # -------------------------------------------------------------------------
    def _upload_processing(self, project_id):
        """Import processing.

        :param project_id: (string)
            Project ID.
        """
        if not has_permission(self._request, 'prj_editor'):
            raise HTTPForbidden()
        upload = self._request.params.get('xml_file')
        if isinstance(upload, basestring):
            return

        tree = load(upload.filename, {'publiforge':
            join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')},
            upload.file.read())
        if isinstance(tree, basestring):
            self._request.session.flash(tree, 'alert')
            return

        error = ProjectProcessing.load(project_id, tree.find('processing'))
        if error is not None:
            self._request.session.flash(error, 'alert')

    # -------------------------------------------------------------------------
    def _settings_form(self, project=None):
        """Return a project settings form.

        :param project: (:class:`~..models.projects.Project` instance,
            optional) Current project object.
        :return: (tuple)
            A tuple such as ``(form, tab_set)``
        """
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=2, max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='description',
            validator=Length(max=DESCRIPTION_LEN), missing=''))
        schema.add(SchemaNode(String(), name='status',
            validator=OneOf(PROJECT_STATUS.keys())))
        schema.add(SchemaNode(Date(), name='deadline', missing=None))
        if project is not None:
            for user in project.users:
                schema.add(SchemaNode(String(), name='usr_%d' % user.user_id,
                validator=OneOf(PROJECT_PERMS.keys())))
            for group in project.groups:
                schema.add(SchemaNode(String(), name='grp_%s' % group.group_id,
                validator=OneOf(PROJECT_PERMS.keys())))

        defaults = {'status': 'draft'}

        return (
            Form(self._request, schema=schema, defaults=defaults, obj=project),
            TabSet(self._request, PROJECT_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _save(self, project, values):
        """Save a project settings.

        :param project: (:class:`~..models.projects.Project` instance)
            Project to update.
        :param values: (dictionary)
            Form values.
        :return: (boolean)
            ``True`` if succeeds.
        """
        i_editor = has_permission(self._request, 'prj_editor')
        for user in project.users:
            if i_editor or user.user_id != self._request.session['user_id']:
                user.perm = values['perm_%d' % user.user_id]

        if 'project' in self._request.session:
            del self._request.session['project']
        if 'build' in self._request.session:
            del self._request.session['build']

        DBSession.commit()
        return True

    # -------------------------------------------------------------------------
    def _add_members(self, project):
        """Add selected users to the project.

        :param project_id: (:class:`~..models.projects.Project` instance)
            Project object.
        """
        if not has_permission(self._request, 'prj_modifier') \
               and self._request.session['project']['perm'] != 'editor':
            return

        user_ids = [k.user_id for k in project.users]
        for user_id in get_action(self._request)[1]:
            user_id = int(user_id)
            if not user_id in user_ids:
                project.users.append(ProjectUser(user_id))
        DBSession.commit()
        del self._request.session['menu']

    # -------------------------------------------------------------------------
    def _add_member_groups(self, project):
        """Add selected groups to the project.

        :param project_id: (:class:`~..models.projects.Project` instance)
            Project object.
        """
        if not has_permission(self._request, 'prj_editor') \
               and self._request.session['project']['perm'] != 'editor':
            return

        group_ids = [k.group_id for k in project.groups]
        for group_id in get_action(self._request)[1]:
            if not group_id in group_ids:
                project.groups.append(ProjectGroup(
                    project.project_id, group_id))
        DBSession.commit()
        del self._request.session['menu']

    # -------------------------------------------------------------------------
    def _download_results(self, build_list, processings, build_ids):
        """Gather results in a ZIP file and return a Pyramid response.

        :param build_list: (list)
            List of builds of current project.
        :param processings: (dictionary)
            Dictionary of used processings.
        :param build_ids: (list)
            List of build IDs to download.
        :return: (:class:`pyramid.response.Response` instance or raise a
            :class:`pyramid.httpexceptions.HTTPNotFound` exception.)
        """
        # Create file list
        filenames = []
        for build in build_list:
            if build['build_id'] in build_ids and build.get('files'):
                path = processings[build['processing_id']][1]
                if current_storage(self._request, path.partition('/')[0]):
                    filenames += [join(path, k) for k in build['files']]

        # Empty list
        if not filenames:
            self._request.session.flash(_('No result to download!'), 'alert')
            return

        # Single file
        storage_root = self._request.registry.settings['storage.root']
        if len(filenames) == 1:
            return file_download(
                self._request, storage_root, filenames, basename(filenames[0]))

        # Several filenames
        tmp = tempfile.NamedTemporaryFile(
            dir=self._request.registry.settings['temporary_dir'])
        zip_file = zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED)
        root = join(storage_root, commonprefix(filenames))
        for name in filenames:
            name = join(storage_root, name)
            zip_file.write(name, relpath(name, root))
        zip_file.close()
        return file_download(self._request, '', (tmp.name,), 'results.zip')



# $Id: views.py ada86c3ea608 2012/03/22 22:52:47 patrick $
"""Some various utilities for views."""

from os import walk
from os.path import exists, join, isfile, dirname, relpath, basename, normpath
import zipfile
import tempfile
from sqlalchemy import select, desc
from colander import SchemaNode, Mapping, String, Integer, Boolean
from colander import All, Length, Regex, OneOf
from webhelpers.html import literal

from pyramid.httpexceptions import HTTPNotFound, HTTPForbidden
from pyramid.i18n import get_localizer
from pyramid.response import Response

from .utils import _, EXCLUDED_FILES, decrypt, get_mime_type, has_permission
from .xml import local_text
from .renderer import Paging
from .form import grid_item
from ..models import VALUE_LEN, DBSession
from ..models.users import  User
from ..models.groups import GROUPS_USERS, Group
from ..models.storages import VCS_ENGINES, Storage, StorageUser, StorageGroup
from ..models.projects import Project, ProjectProcessing, ProjectProcessingFile
from ..models.projects import ProjectUser, ProjectGroup
from ..models.projects import ProjectPack, ProjectPackFile


# =============================================================================
def get_action(request):
    """Return a tuple such as ``(action, items)`` where ``action`` is a
    string such as ``<act><?|!><item_id>`` and ``items`` is a list of
    selected items in a list form.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :return: (tuple)
        A tuple such as ``(action, items)``.

    Each submit button returns a string such as ``<act><?|!><item_id>.x`` where
    ``<item_id>`` is the item identifier or ``#`` for all selected items,
    ``<?|!>`` means respectively *confirm* or *proceed* and ``<act>`` is one of
    the following action:

    * ``agp``: add groups
    * ``aur``: add users
    * ``bld``: build
    * ``ccl``: cancel
    * ``cls``: close
    * ``del``: delete
    * ``des``: description
    * ``dif``: get diff
    * ``dir``: make directory
    * ``dnl``: download
    * ``exp``: export
    * ``imp``: import
    * ``npk``: create a new pack with selection
    * ``pck``: add to pack
    * ``prc``: add to processing
    * ``ren``: rename
    * ``rgp``: remove groups
    * ``rur``: remove users
    * ``rmv``: remove
    * ``sch``: search
    * ``shw``: show
    * ``syn``: synchronize
    * ``upl``: upload

    Checkbox inputs return string such as ``#<item_id>``.

    For instance, ``del!#`` and ``['#user1', '#user2']`` means "delete
    ``user1`` and ``user2``". ``syn!DataHg`` means "synchronize
    ``DataHg`` and only this one".
    """
    action = ''
    items = []
    for param in request.params:
        if param[0] == '#':
            items.append(param[1:])
        elif param[-2:] == '.x':
            if param[-3:-2] == '#':
                action = param[0:-2]
            else:
                return param[0:-2], (param[4:-2],)
    if '#' in action and not items:
        request.session.flash(_('Select items!'), 'alert')
        action = ''
    return action, items


# =============================================================================
def paging_users(request, only_active=True):
    """Return a :class:`~.renderer.Paging` object filled with users.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param only_active: (boolean, default=True)
        if ``True``, only active users are included.
    :return: (tuple)
        A tuple such as ``(paging, filters)`` where ``paging`` is a
        :class:`~.renderer.Paging` object and ``filters`` a dictionary of
        filters.
    """
    # Parameters
    paging_id = only_active and 'users' or 'users!'
    params = Paging.params(request, paging_id, '+login')
    if only_active or (len(request.POST) == 0 and not 'f_status' in params):
        params['f_status'] = 'active'

    # Query
    query = DBSession.query(User.user_id, User.login, User.name, User.status)
    if 'f_login' in params:
        query = query.filter(
            User.login.like('%%%s%%' % params['f_login'].lower()))
    if 'f_name' in params:
        query = query.filter(User.name.ilike('%%%s%%' % params['f_name']))
    if 'f_status' in params and params['f_status'] != '*':
        query = query.filter(User.status == params['f_status'])
    if 'f_group' in params:
        query = query.join(GROUPS_USERS).filter(
            GROUPS_USERS.c.group_id == params['f_group'])

    # Order by
    oby = 'users.%s' % params['sort'][1:]
    query = query.order_by(desc(oby) if params['sort'][0] == '-' else oby)

    return Paging(request, paging_id, query), params


# =============================================================================
def paging_groups(request, paging_id='groups'):
    """Return a :class:`~.renderer.Paging` object filled with groups.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param paging_id: (string, default='groups')
        Paging ID: ``'groups'`` for adding purpose and ``groups!`` for
        administration purpose.
    :return: (tuple)
        A tuple such as ``(paging, filters)`` where ``paging`` is a
        :class:`~.renderer.Paging` object and ``filters`` a dictionary of
        filters.
    """
    # Parameters
    params = Paging.params(request, paging_id, '+label')

    # Query
    query = DBSession.query(Group)
    if 'f_label' in params:
        query = query.filter(Group.label.ilike('%%%s%%' % params['f_label']))

    # Order by
    oby = 'groups.%s' % params['sort'][1:]
    query = query.order_by(desc(oby) if params['sort'][0] == '-' else oby)

    return Paging(request, paging_id, query), params


# =============================================================================
def current_storage(request, storage_id=None, only_dict=True):
    """Initialize, if necessary, ``request.session['storage']`` and
    return it as current storage dictionary.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param storage_id: (integer, optional)
        ID of storage to get context.
    :param only_dict: (boolean, default=True)
        If ``False``, it returns the entire
        :class:`~.models.storages.Storage` object.
    :return: (tuple)
        A tuple such as ``(storage_dictionary, handler)`` or a
        :class:`pyramid.httpexceptions.`HTTPNotFound` exception.

    ``storage_dictionary`` has the following keys: ``storage_id``,
    ``label``, ``description``, ``perm``, ``vcs_engine``, ``vcs_url``,
    ``public_url``.
    """
    # Already in session
    storage_id = storage_id or request.matchdict['storage_id']
    if only_dict and 'storage' in request.session and request\
           .session['storage']['storage_id'] == storage_id:
        handler = request.registry['handler'].get_handler(storage_id)
        if handler is not None:
            return request.session['storage'], handler

    # Read in database
    storage = DBSession.query(Storage).filter_by(
        storage_id=storage_id).first()
    if storage is None:
        raise HTTPNotFound(comment=_('This storage does not exist!'))

    # Permission
    user_id = request.session['user_id']
    perm = (has_permission(request, 'stg_editor') or storage.access == 'open')\
           and 'editor'
    if perm != 'editor':
        record = DBSession.query(StorageUser).filter_by(
            storage_id=storage_id, user_id=user_id).first()
        perm = record and record.perm
    if perm != 'editor':
        groups = [k.group_id for k in DBSession.execute(
            select([GROUPS_USERS], GROUPS_USERS.c.user_id == user_id))]
        if groups:
            record = [k[0] for k in DBSession.query(StorageGroup.perm)
                .filter(StorageGroup.storage_id == storage_id)
                .filter(StorageGroup.group_id.in_(groups))]
            perm = ('editor' in record and 'editor') or (record and record[0])\
                   or perm
    if perm is None:
        raise HTTPForbidden(comment=_(
            'You do not have access to this storage!'))

    request.session['storage'] = {
        'storage_id': storage_id,
        'label': storage.label,
        'description': storage.description,
        'perm': perm,
        'vcs_engine': VCS_ENGINES[storage.vcs_engine],
        'vcs_url': storage.vcs_url,
        'public_url': storage.public_url}
    handler = request.registry['handler'].get_handler(storage_id, storage)
    return only_dict and (request.session['storage'], handler) \
           or (storage, handler)


# =============================================================================
def vcs_user(request, storage):
    """Return a VCS user tuple.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param storage: (dictionary)
        Storage dictionary.
    :return: (tuple)
        A VCS user tuple ``(user_id, password, name)``.
    """
    user = DBSession.query(StorageUser).filter_by(
        storage_id=storage['storage_id'],
        user_id=request.session['user_id']).first()
    if user is None:
        return None, None, request.session['name']
    password = decrypt(
        user.vcs_password, request.registry.settings['auth.key'])
    return user.vcs_user, password, request.session['name']


# =============================================================================
def file_infos(request, record):
    """Return a dictionary of file informations.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param record: (:class:`~.models.projects.ProjectProcessing`
        or :class:`~.models.projects.ProjectPack` instance).
    :return: (dictionary)
        A dictionary with keys ``file``, ``resource`` and ``template``. Each
        value of this dictionary is a tuple of tuples such as ``(file_type,
        storage_id, path, target, visible)``.
    """
    root_path = request.registry.settings['storage.root']
    files = {'file': [], 'resource': [], 'template': []}
    for item in record.files:
        files[item.file_type].append((
            get_mime_type(join(root_path, item.path))[1],
            item.path.partition('/')[0], item.path.partition('/')[2],
            item.target, hasattr(item, 'visible') and item.visible))
    return files


# =============================================================================
def file_download(request, path, filenames, download_name=None):
    """Prepare file for download and return a Pyramid response.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param path: (string)
        Absolute path to files.
    :param filenames: (list)
        List of files to download.
    :param download_name: (string)
        Visible name during download.
    :return: (:class:`pyramid.response.Response` instance or raise a
        :class:`pyramid.httpexceptions.HTTPNotFound` exception.)
    """
    # Single file
    if len(filenames) == 1 and isfile(join(path, filenames[0])):
        fullname = join(path, filenames[0])
        if not exists(fullname):
            raise HTTPNotFound(comment=_('This file does not exist!'))
        with open(fullname, 'r') as hdl:
            content = hdl.read()
        download_name = download_name or filenames[0]
        response = Response(content, content_type=get_mime_type(fullname)[0])
        response.headerlist.append(('Content-Disposition',
            'attachment; filename="%s"' % download_name.encode('utf8')))
        return response

    # Several files
    tmp = tempfile.NamedTemporaryFile(
        dir=request.registry.settings['temporary_dir'])
    zip_file = zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED)
    for name in filenames:
        fullname = normpath(join(path, name))
        if not fullname.startswith(path) or not exists(fullname):
            zip_file.close()
            raise HTTPNotFound(comment=_('This file does not exist!'))
        if isfile(fullname):
            zip_file.write(fullname, name)
            continue
        for root, dirs, files in walk(fullname):
            for name in dirs:
                if name in EXCLUDED_FILES:
                    dirs.remove(name)
            for name in files:
                if not name in EXCLUDED_FILES:
                    name = join(root, name)
                    try:
                        zip_file.write(name, relpath(name, path))
                    except IOError, err:
                        raise HTTPNotFound(comment=err)
    zip_file.close()
    if not download_name:
        download_name = \
            filenames[0] if len(filenames) == 1 else basename(path)
    return file_download(request, '', (tmp.name,), '%s.zip' % download_name)


# =============================================================================
def create_pack(request, filenames, path='.'):
    """Create a new pack with selected files.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param filenames: (list)
        Names of files to add to new pack.
    :param path: (string, optional)
        Common path.
    :return: (tuple)
        A tuple such as ``(project_id, pack_id``) or ``(None, None)`` if
        failed.
    """
    label = ', '.join([basename(k) for k in filenames])
    project_id = request.session['project']['project_id']
    if DBSession.query(ProjectPack) \
           .filter_by(project_id=project_id, label=label).first():
        request.session.flash(_('This pack already exists.'), 'alert')
        return None, None

    pack = ProjectPack(project_id, label)
    for name in filenames:
        pack.files.append(
            ProjectPackFile('file', normpath(join(path, name))))
        DBSession.add(pack)
        DBSession.commit()
    return pack.project_id, pack.pack_id


# =============================================================================
def add2container(request, container_type, file_type, path, filenames):
    """Add files to a processing or a pack.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param container_type: ('prc' or 'pck')
        ``'prc'`` for processing and ``'pck'`` for pack.
    :param file_type: ('fil', 'rsc', 'tpl', 'out')
        ``'fil'`` for file,``'rsc'`` for resource, ``'tpl'`` for template and
        ``'out'`` for output directory.
    :param path:
        Relative path from storage root.
    :param filenames: (list)
        Names of files to add.
    """
    if not 'container' in request.session:
        return
    project_id = request.session['container']['project_id']
    file_type = {'fil': 'file', 'rsc': 'resource', 'tpl': 'template',
                 'out': 'output'}[file_type]
    # Output
    if file_type == 'output':
        if len(filenames) > 1:
            request.session.flash(_('Select only one directory!'), 'alert')
            return
        name = join(
            request.registry.settings['storage.root'], path, filenames[0])
        if isfile(name):
            filenames[0] = dirname(filenames[0])
        record = DBSession.query(ProjectProcessing).filter_by(
            project_id=project_id, processing_id=request
            .session['container']['processing_id']).first()
        if record is not None:
            record.output = normpath(join(path, filenames[0]))
            DBSession.commit()
            request.session.flash(_(
                'Output directory has been changed to "${n}".',
                {'n': record.output}), 'message')
        return

    # Container
    if container_type == 'pck':
        record = DBSession.query(ProjectPack).filter_by(
            project_id=project_id,
            pack_id=request.session['container']['pack_id']).first()
        container_class = ProjectPackFile
    else:
        record = DBSession.query(ProjectProcessing).filter_by(
            processing_id=request.session['container']['processing_id'],
            project_id=project_id).first()
        container_class = ProjectProcessingFile
    if record is None:
        return
    paths = [k.path for k in record.files if k.file_type == file_type]

    # Browse files
    added = []
    for name in filenames:
        name = normpath(join(path, name))
        if name in paths:
            continue
        record.files.append(container_class(file_type, name))
        added.append(name)
    DBSession.commit()

    # Message
    if len(added) > 0:
        record = {'a': ', '.join(added)}
        added = {'file': _('Files to process added: ${a}', record),
                 'resource': _('Resource files added: ${a}', record),
                 'template': _('Template files added: ${a}', record)
                 }[file_type]
        request.session.flash(added, 'message')


# =============================================================================
def current_project(request, project_id=None, only_dict=True):
    """Update ``request.session['project']`` and return it as current project
    dictionary or return a SqlAlchemy object.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param project_id: (integer, optional)
        ID of project to get context.
    :param only_dict: (boolean, default=True)
        If ``False``, it returns the entire :class:`~.models.projects.Project`
        object.
    :return: (dictionary or ``None``)
        ``None`` if error or a dictionary with the following keys:
        ``project_id``, ``label``, ``description``, ``perm``, ``deadline``,
        ``processing_id`` (default processing ID) and ``processing_labels``.
    """
    # Already in session
    project_id = project_id or request.matchdict.get('project_id')
    if project_id is None:
        raise HTTPNotFound(comment=_('You must specify a project ID!'))
    project_id = int(project_id)

    if only_dict and 'project' in request.session \
           and request.session['project']['project_id'] == project_id:
        if 'processing_id' in request.params:
            request.session['project']['processing_id'] = \
                int(request.params['processing_id'])
        return request.session['project']

    # Read in database
    project = DBSession.query(Project).filter_by(project_id=project_id).first()
    if project is None:
        raise HTTPNotFound(comment=_('This project does not exist!'))
    processing_labels = DBSession.query(
        ProjectProcessing.processing_id, ProjectProcessing.label)\
        .filter_by(project_id=project_id)\
        .order_by(ProjectProcessing.label).all()

    # Permission?
    user_id = request.session['user_id']
    perm = has_permission(request, 'prj_editor') and 'editor'
    if perm != 'editor':
        record = DBSession.query(ProjectUser).filter_by(
            project_id=project_id, user_id=user_id).first()
        perm = record and record.perm
    if perm != 'editor':
        groups = [k.group_id for k in DBSession.execute(
            select([GROUPS_USERS], GROUPS_USERS.c.user_id == user_id))]
        if groups:
            record = [k[0] for k in DBSession.query(ProjectGroup.perm)
                .filter(ProjectGroup.project_id == project_id)
                .filter(ProjectGroup.group_id.in_(groups))]
            perm = ('editor' in record and 'editor') or (record and record[0])\
                   or perm
    if perm is None:
        raise HTTPForbidden(comment=_('You do not work in this project!'))

    request.session['project'] = {'project_id': project_id,
        'label': project.label, 'description': project.description,
        'perm': perm,
        'deadline': project.deadline,
        'processing_id': int(request.params.get('processing', 1)),
        'processing_labels': processing_labels}
    return only_dict and request.session['project'] or project


# =============================================================================
def current_processing(request, project_id=None, processing_id=None,
                       container=False):
    """Return the current processing object.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param project_id: (integer, optional)
        ID of project to retrieve processing.
    :param processing_id: (integer, optional)
        ID of processing to retrieve.
    :param container: (boolean)
        If ``True``, add current processing container.
    :return: (tuple)
        A tuple such as ``(processing, processor)``.

    ``processing`` is a :class:`~.models.projects.ProjectProcessing`
    object. ``processor`` is a :class:`lxml.etree.ElementTree` object
    representing the used processor. If this function fails, it raises a
    :class:`pyramid.httpexceptions.`HTTPNotFound` exception.

    ``request.session['container']`` is updated with ``processing_id``,
    ``processing_label`` and ``has_output``.
    """
    # IDs
    project_id = project_id or request.matchdict.get('project_id')
    if project_id is None:
        raise HTTPNotFound(comment=_('You must specify a project ID!'))
    processing_id = processing_id or request.matchdict.get('processing_id')
    if processing_id is None:
        raise HTTPNotFound(comment=_('You must specify a processing ID!'))
    project_id = int(project_id)
    processing_id = int(processing_id)

    # Processing
    processing = DBSession.query(ProjectProcessing).filter_by(
        project_id=project_id, processing_id=processing_id).first()
    if processing is None:
        raise HTTPNotFound(comment=_('This processing does not exist!'))

    # Container for storage
    if container:
        request.session['container'] = {
            'project_id': project_id,
            'processing_id': processing_id,
            'processing_label': processing.label,
            'has_output': bool(processing.output)}

    # Processor
    processor = request.registry['fbuild'].processor(
        request, processing.processor)
    if processor is None:
        raise HTTPNotFound(comment=_('This processor does not exist!'))

    return processing, processor


# =============================================================================
def variable_schema(processor, variables, only_visible=False,
                    with_default=True):
    """Return a schema to validate variable form.

    :param processor: (:class:`lxml.etree.ElementTree` instance)
        Processor of current processing.
    :param variables: (dictionary)
        Variables of current processing.
    :param only_visible: (boolean, default=False)
        If ``True``, process only visible variables.
    :param with_default: (boolean, default=True)
        If ``True``, create schema for default values.
    :return: (tuple)
        A tuple such as ``(schema, defaults)`` where ``schema`` is a
        :class:`colander.SchemaNode` instance and ``defaults`` a dictionary.
    """
    def node(name, var):
        """Return a SchemaNode according to ``var`` type."""
        var_type = var.get('type')
        if var_type == 'boolean':
            node = SchemaNode(Boolean(), name=name, missing=False)
        elif var_type == 'integer':
            node = SchemaNode(Integer(), name=name, missing=0)
        elif var_type == 'select':
            options = [k.get('value', k.text) for k in var.findall('option')]
            node = SchemaNode(String(), name=name, validator=OneOf(options))
        elif var_type == 'regex':
            node = SchemaNode(String(), name=name, validator=All(
                Regex(var.findtext('pattern').strip()), Length(max=VALUE_LEN)))
        else:
            node = SchemaNode(String(), name=name,
                validator=Length(max=VALUE_LEN), missing='')
        return node

    schema = SchemaNode(Mapping())
    form_defaults = {}
    for var in processor.findall('processor/variables/group/var'):
        name = var.get('name')
        if only_visible \
               and (not name in variables or not variables[name].visible):
            continue

        schema.add(node(name, var))
        form_defaults[name] = variables[name].value if name in variables \
                              else var.findtext('default') is not None \
                              and var.findtext('default').strip() or ''
        if var.get('type') == 'boolean':
            form_defaults[name] = (form_defaults[name] == 'true')

        if not only_visible:
            schema.add(
                SchemaNode(Boolean(), name='%s_see' % name, missing=False))
            form_defaults['%s_see' % name] = \
                (name in variables and variables[name].visible) \
                or var.get('visible') or False

        if with_default:
            schema.add(node('%s_def' % name, var))
            form_defaults['%s_def' % name] = \
                variables[name].default if name in variables \
                else var.findtext('default') is not None \
                and var.findtext('default').strip() or ''
            if var.get('type') == 'boolean':
                form_defaults['%s_def' % name] = (
                    form_defaults['%s_def' % name] == 'true')

    return schema, form_defaults


# =============================================================================
def describe_var(request, processor, variables, name,
                 show_processor_default=False):
    """Return a description of variable ``name``.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param processor: (:class:`lxml.etree.ElementTree` instance)
        Processor of current processing.
    :param variables: (dictionary)
        Variables of current processing.
    :param name: (string)
        Name of the variable.
    :param show_processor_default: (boolean, default=False)
        Show the processor default value if any.
    :return: (string)
    """
    var = processor.xpath('processor/variables/group/var[@name="%s"]' % name)
    if not len(var):
        return
    var = var[0]
    translate = get_localizer(request).translate

    html = grid_item(
        '', translate(_('Name:')), literal('<b>%s</b>' % name), class_='')
    html += grid_item(
        '', translate(_('Type:')), var.get('type'), class_='')
    if name in variables:
        html += grid_item(
            '', translate(_('Value:')), variables[name].value, class_='')
        html += grid_item(
            '', translate(_('Suggested:')), variables[name].default, class_='')
    if show_processor_default and var.findtext('default') is not None:
        html += grid_item(
            '', translate(_('Default:')), var.findtext('default'), class_='')

    description = ''
    if var.find('description') is not None:
        description += local_text(var, 'description', request)\
                       .replace(' --', '<br/>')
    if var.getparent().find('description') is not None:
        description += '<br/><br/>'
        description += local_text(var.getparent(), 'description', request) \
                     .replace(' --', '<br/>')
    if description:
        html += grid_item(
            '', translate(_('Description:')), literal(description), class_='')

    return html



# $Id: storage.py ada86c3ea608 2012/03/22 22:52:47 patrick $
# pylint: disable = I0011, C0322
"""Storage view callables."""

from os.path import join, isdir, dirname
from webhelpers.html import HTML
from sqlalchemy import select, desc, or_, and_
from colander import Mapping, SchemaNode, String, Integer, Length, OneOf

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound, HTTPForbidden

from ..lib.utils import _, FILE_TYPES, has_permission, encrypt, age
from ..lib.xml import upload_configuration, export_configuration
from ..lib.views import vcs_user, paging_users, paging_groups, current_storage
from ..lib.views import get_action, file_download, create_pack, add2container
from ..lib.form import Form
from ..lib.renderer import TabSet, sortable_column, Paging
from ..models import ID_LEN, LABEL_LEN, DESCRIPTION_LEN, PATH_LEN
from ..models import DBSession, close_dbsession
from ..models.users import User
from ..models.groups import GROUPS_USERS, Group
from ..models.storages import STORAGE_ACCESS, VCS_ENGINES, STORAGE_PERMS
from ..models.storages import Storage, StorageUser, StorageGroup


STORAGE_SETTINGS_TABS = (
    _('Description'), _('Authorized users'), _('Authorized groups'))
REFRESH = {
    900: _('every 15 minutes'), 1800: _('every 30 minutes'),
    3600: _('every hour'), 7200: _('every 2 hours'),
    14400: _('every 4 hours'), 28800: _('every 8 hours'),
    86400: _('every day')}


# =============================================================================
class StorageView(object):
    """Class to manage storages."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request
        self._vcs_engines = dict([(k, VCS_ENGINES[k]) for k in
            request.registry.settings['storage.available_vcs'].split()])

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_admin',
                 renderer='../Templates/stg_admin.pt',
                 permission='stg.update')
    def admin(self):
        """List storages for administration purpose."""
        action, items = get_action(self._request)
        if action == 'imp!':
            upload_configuration(self._request, 'stg_manager', 'storage')
            action = ''
            del self._request.session['menu']
        elif action[0:4] == 'del!':
            self._delete_storages(items)
            action = ''
            del self._request.session['menu']
        elif action[0:4] == 'exp!':
            return self._export_storages(items)

        paging, defaults = self._paging_storages()
        form = Form(self._request, defaults=defaults)
        groups = [(k.group_id, k.label) for k in
                  DBSession.query(Group).join(StorageGroup)]

        depth = (self._request.breadcrumbs.current_path() ==
                 self._request.route_path('site_admin') and 3) or 2
        self._request.breadcrumbs.add(_('Storage administration'), depth)
        return {
            'form': form, 'paging': paging, 'action': action,
            'groups': groups, 'STORAGE_ACCESS': STORAGE_ACCESS,
            'vcs_engines': self._vcs_engines,
            'i_editor': has_permission(self._request, 'stg_editor'),
            'i_manager': has_permission(self._request, 'stg_manager')}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_index',
                 renderer='../Templates/stg_index.pt',
                 permission='stg.read')
    def index(self):
        """List authorized storages."""
        # Action
        user_id = self._request.session['user_id']
        action, items = get_action(self._request)
        if action[0:4] == 'exp!':
            return self._export_storages(items)
        elif action[0:3] == 'men':
            stg_user = DBSession.query(StorageUser).filter_by(
                storage_id=action[4:], user_id=user_id).first()
            if stg_user is None:
                stg_user = StorageUser(action[4:], user_id)
                DBSession.add(stg_user)
            stg_user.in_menu = action[3] == '+'
            DBSession.commit()
            del self._request.session['menu']
        elif action == 'cls!' and 'container' in self._request.session:
            del self._request.session['container']

        # Form and environment
        paging, defaults = self._paging_storages(user_id)
        form = Form(self._request, defaults=defaults)
        changes = {}
        for storage in paging:
            handler = self._request.registry['handler']\
                  .get_handler(storage.storage_id, storage)
            changes[handler.uid] = list(handler.vcs.last_change())
            changes[handler.uid][0] = age(changes[handler.uid][0])
            force = (action == 'syn!%s' % storage.storage_id) or \
                    (action == 'syn!#' and storage.storage_id in items)
            if handler.synchronize(self._request, force):
                handler.cache.clear()
        in_menus = [k[0] for k in DBSession.query(StorageUser.storage_id)
                    .filter_by(user_id=self._request.session['user_id'])
                    .filter_by(in_menu=True).all()]

        # Working?
        action, progress = \
            self._request.registry['handler'].progress(changes.keys())
        if action and not 'ajax' in self._request.params:
            self._request.response.headerlist.append(('Refresh',
                self._request.registry.settings.get('refresh.long', '8')))

        self._request.breadcrumbs.add(_('All storages'),
            'container' in self._request.session and 6 or 2)
        return {
            'form': form, 'paging': paging, 'progress': progress,
            'vcs_engines': self._vcs_engines, 'changes': changes,
            'in_menus': in_menus}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_index_ajax', renderer='json',
                 permission='stg.read')
    def index_ajax(self):
        """Authorized storage status for AJAX request."""
        user_id = self._request.session['user_id']
        paging = self._paging_storages(user_id)[0]
        working, progress = self._request.registry['handler'].progress(
            [k.storage_id for k in paging])
        response = dict([(k, progress[k][0]) for k in progress])
        response['working'] = working
        return response

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_view',
                 renderer='../Templates/stg_view.pt',
                 permission='stg.read')
    def view(self):
        """Show storage settings with its users and groups."""
        action = get_action(self._request)[0]
        if action == 'exp!':
            return self._export_storages((
                self._request.matchdict.get('storage_id'),))

        storage = current_storage(self._request, only_dict=False)[0]
        tab_set = TabSet(self._request, STORAGE_SETTINGS_TABS)
        members = DBSession.query(
            User.user_id, User.login, User.name, StorageUser.perm)\
            .join(StorageUser)\
            .filter(StorageUser.storage_id == storage.storage_id)\
            .filter(StorageUser.perm != None).order_by(User.login)
        member_groups = DBSession.query(
            Group.group_id, Group.label, StorageGroup.perm)\
            .filter(StorageGroup.storage_id == storage.storage_id)\
            .filter(Group.group_id == StorageGroup.group_id)\
            .order_by(Group.label).all()
        i_editor = has_permission(self._request, 'stg_editor') \
            or self._request.session['storage']['perm'] == 'editor'

        self._request.breadcrumbs.add(
            _('Storage settings'), replace=self._request.route_path(
            'storage_edit', storage_id=storage.storage_id))
        return {
            'form': Form(self._request), 'tab_set': tab_set,
            'STORAGE_ACCESS': STORAGE_ACCESS, 'REFRESH': REFRESH,
            'perms': STORAGE_PERMS, 'vcs_engines': self._vcs_engines,
            'storage': storage, 'members': members,
            'member_groups': member_groups, 'i_editor': i_editor}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_create',
                 renderer='../Templates/stg_edit.pt',
                 permission='stg.create')
    def create(self):
        """Create a storage."""
        form, tab_set = self._settings_form()

        if form.validate():
            storage = self._create(form.values)
            if storage is not None:
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path(
                    'storage_edit', storage_id=storage.storage_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        self._request.breadcrumbs.add(_('Storage creation'))
        return {
            'form': form, 'tab_set': tab_set,
            'STORAGE_ACCESS': STORAGE_ACCESS, 'REFRESH': REFRESH,
            'perms': STORAGE_PERMS, 'vcs_engines': self._vcs_engines,
            'storage': None, 'members': None, 'group_members': None,
            'paging': None, 'groups': None}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_edit',
                 renderer='../Templates/stg_edit.pt')
    def edit(self):
        """Edit storage settings."""
        # Authorization
        storage = current_storage(self._request, only_dict=False)[0]
        if not has_permission(self._request, 'stg_editor') \
               and self._request.session['storage']['perm'] != 'editor':
            raise HTTPForbidden()

        # Action
        paging = groups = None
        action = get_action(self._request)[0]
        if action == 'aur?' or \
               (not action and self._request.GET.get('paging_id') == 'users'):
            groups = DBSession.query(Group.group_id, Group.label).all()
            paging = paging_users(self._request)[0]
        elif action[0:4] == 'aur!':
            self._add_members(storage)
        elif action[0:4] == 'rur!':
            DBSession.query(StorageUser).filter_by(
                storage_id=storage.storage_id, user_id=int(action[4:]))\
                .delete()
            DBSession.commit()
            del self._request.session['menu']
        elif action == 'agp?' or (not action and
                self._request.GET.get('paging_id') == 'groups'):
            paging = paging_groups(self._request)[0]
        elif action[0:4] == 'agp!':
            self._add_member_groups(storage)
        elif action[0:4] == 'rgp!':
            DBSession.query(StorageUser).filter_by(
                storage_id=storage.storage_id, user_id=int(action[4:]),
                perm=None).delete()
            DBSession.query(StorageGroup).filter_by(
                storage_id=storage.storage_id, group_id=action[4:]).delete()
            DBSession.commit()
            del self._request.session['menu']

        # Environment
        form, tab_set = self._settings_form(storage)
        members = DBSession.query(
            User.user_id, User.login, User.name, StorageUser.perm)\
            .filter(StorageUser.storage_id == storage.storage_id)\
            .filter(User.user_id == StorageUser.user_id)\
            .order_by(User.login).all()
        member_groups = DBSession.query(
            Group.group_id, Group.label, StorageGroup.perm)\
            .filter(StorageGroup.storage_id == storage.storage_id)\
            .filter(Group.group_id == StorageGroup.group_id)\
            .order_by(Group.label).all()

        # Save
        view_path = self._request.route_path('storage_view',
            storage_id=storage.storage_id)
        if action == 'sav!' and form.validate(storage):
            storage.set_vcs_password(self._request.registry.settings,
                form.values.get('password'))
            for user in storage.users:
                user.perm = form.values['usr_%d' % user.user_id]
            for group in storage.groups:
                group.perm = form.values['grp_%s' % group.group_id]
            DBSession.commit()
            return HTTPFound(view_path)
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs.add(
            _('Storage settings'), replace=view_path)

        return {
            'form': form, 'action': action, 'tab_set': tab_set,
            'STORAGE_ACCESS': STORAGE_ACCESS, 'REFRESH': REFRESH,
            'perms': STORAGE_PERMS, 'vcs_engines': self._vcs_engines,
            'storage': storage, 'members': members,
            'member_groups': member_groups, 'paging': paging, 'groups': groups}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_root',
                 renderer='../Templates/stg_browse.pt',
                 permission='stg.read')
    @view_config(route_name='storage_browse',
                 renderer='../Templates/stg_browse.pt',
                 permission='stg.read')
    def browse(self):
        """Browse a storage."""
        # Environment
        storage, handler = current_storage(self._request)
        storage_id = storage['storage_id']
        path = self._request.matchdict.get('path', [])
        path, real_path, html_path = self._html_path(storage_id, path)

        # Form, action & working status
        form, action, items, working = \
            self._browse_action(storage, handler, path)
        if action[0:4] == 'dnl!':
            return file_download(self._request, real_path, items)
        elif action[0:4] == 'sch!':
            return HTTPFound(self._request.route_path('file_search'))
        elif action[0:4] == 'npk!' and items[0] is not None:
            return HTTPFound(self._request.route_path(
                'pack_edit', project_id=items[0], pack_id=items[1]))

        # Directory content
        dirs, files = handler.dir_infos(
            path, self._request.params.get('sort', '+name'))

        # Working?
        working, progress = self._request\
            .registry['handler'].progress((storage_id,), working)
        if working and not 'ajax' in self._request.params:
            self._request.response.headerlist.append(('Refresh',
                self._request.registry.settings.get('refresh.short', '3')))
        if working:
            handler.cache.clear()
        if storage_id in progress and progress[storage_id][0] == 'error':
            self._request.session.flash(progress[storage_id][1], 'alert')

        self._request.breadcrumbs.add(_('Storage browsing'), root_chunks=3)
        return {
            'form': form, 'action': action, 'working': working,
            'sortable_column': sortable_column, 'storage': storage,
            'path': path, 'html_path': html_path, 'dirs': dirs,
            'files': files, 'FILE_TYPES': FILE_TYPES,
            'vcs_engines': self._vcs_engines}

    # -------------------------------------------------------------------------
    @view_config(route_name='storage_root_ajax', renderer='json',
                 permission='stg.read')
    @view_config(route_name='storage_browse_ajax', renderer='json',
                 permission='stg.read')
    def browse_ajax(self):
        """Browse a storage for AJAX request."""
        storage_id = self._request.matchdict['storage_id']
        working = self._request.registry['handler'].progress((storage_id,))[0]
        return {'working': working}

    # -------------------------------------------------------------------------
    def _paging_storages(self, user_id=None):
        """Return a :class:`~..lib.renderer.Paging` object filled
        with storages.

        :param user_id: (integer, optional)
            Select only storages of user ``user_id``.
        :return: (tuple)
            A tuple such as ``(paging, filters)`` where ``paging`` is a
            :class:`~..lib.renderer.Paging` object and ``filters`` a
            dictionary  of filters.
        """
        # Parameters
        paging_id = user_id is None and 'storages!' or 'storages'
        params = Paging.params(self._request, paging_id, '+storage_id')
        if user_id is not None and 'f_access' in params:
            del params['f_access']

        # Query
        query = DBSession.query(Storage)
        if user_id is not None:
            groups = [k.group_id for k in DBSession.execute(
                select([GROUPS_USERS], GROUPS_USERS.c.user_id == user_id))]
            if groups:
                query = query.filter(or_(
                    Storage.access == 'open',
                    and_(StorageUser.user_id == user_id,
                         StorageUser.perm != None,
                         StorageUser.storage_id == Storage.storage_id),
                    and_(StorageGroup.group_id.in_(groups),
                         StorageGroup.storage_id == Storage.storage_id))
                    ).distinct(Storage.storage_id, Storage.label)
            else:
                query = query.join(StorageUser).filter(or_(
                    Storage.access == 'open',
                    and_(StorageUser.user_id == user_id,
                         StorageUser.perm != None)))
        elif 'f_login' in params:
            query = query.join(StorageUser).join(User)\
                    .filter(User.login == params['f_login'])\
                    .filter(StorageUser.perm != None)
        if 'f_id' in params:
            query = query.filter(Storage.storage_id.ilike(
                '%%%s%%' % params['f_id']))
        if 'f_group' in params:
            query = query.join(StorageGroup).filter(
                StorageGroup.group_id == params['f_group'])
        if 'f_label' in params:
            query = query.filter(
                Storage.label.ilike('%%%s%%' % params['f_label']))
        if 'f_access' in params and params['f_access'] != '*':
            query = query.filter(Storage.access == params['f_access'])
        elif not 'f_access' in params:
            query = query.filter(Storage.access != 'closed')
        if 'f_engine' in params:
            query = query.filter(Storage.vcs_engine == params['f_engine'])

        # Order by
        oby = 'storages.%s' % params['sort'][1:]
        query = query.order_by(desc(oby) if params['sort'][0] == '-' else oby)

        return Paging(self._request, paging_id, query), params

    # -------------------------------------------------------------------------
    def _delete_storages(self, storage_ids):
        """Delete storages.

        :param storage_ids: (list)
            List of storage IDs to delete.
        """
        if not has_permission(self._request, 'stg_manager'):
            raise HTTPForbidden()
        storage_root = self._request.registry.settings['storage.root']
        for storage_id in storage_ids:
            Storage.delete(storage_root, storage_id)

    # -------------------------------------------------------------------------
    def _export_storages(self, storage_ids):
        """Export storages.

        :param storage_ids: (list)
            List of storage IDs to export.
        :return: (:class:`pyramid.response.Response` instance)
        """
        i_editor = has_permission(self._request, 'stg_editor')
        user_id = self._request.session['user_id']
        groups = set([k.group_id for k in DBSession.execute(
            select([GROUPS_USERS], GROUPS_USERS.c.user_id == user_id))])
        elements = []
        for storage in DBSession.query(Storage)\
                .filter(Storage.storage_id.in_(storage_ids))\
                .order_by('label'):
            if i_editor or storage.access == 'open' \
                   or user_id in [k.user_id for k in storage.users] \
                   or (groups and
                       groups & set([k.group_id for k in storage.groups])):
                elements.append(storage.xml())

        name = '%s_storages.pfs' % self._request.registry.settings.get(
            'skin.label', 'publiforge')
        return export_configuration(elements, name)

    # -------------------------------------------------------------------------
    def _settings_form(self, storage=None):
        """Return a storage settings form.

        :param storage: (:class:`~..models.storages.Storage` instance,
            optional) Current storage object.
        :return: (tuple)
            A tuple such as ``(form, tab_set)``.
        """
        vcs_engine = storage and storage.vcs_engine \
                     or self._request.params.get('vcs_engine')

        # Schema
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='access',
            validator=OneOf(STORAGE_ACCESS.keys())))
        if storage is None:
            schema.add(SchemaNode(String(), name='storage_id',
                validator=Length(max=ID_LEN)))
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=2, max=LABEL_LEN)))
        schema.add(SchemaNode(String(), name='description',
            validator=Length(max=DESCRIPTION_LEN), missing=''))
        if storage is None:
            schema.add(SchemaNode(String(), name='vcs_engine',
                validator=OneOf(self._vcs_engines.keys())))
        if vcs_engine is None or vcs_engine not in ('none', 'local'):
            schema.add(SchemaNode(String(), name='vcs_url',
                validator=Length(max=PATH_LEN)))
            schema.add(SchemaNode(String(), name='vcs_user',
                validator=Length(max=ID_LEN)))
            schema.add(SchemaNode(String(), name='password', missing=None))
        if vcs_engine is None or vcs_engine in ('none', 'local'):
            schema.add(SchemaNode(String(), name='public_url',
                validator=Length(max=PATH_LEN), missing=None))
        schema.add(SchemaNode(Integer(), name='refresh', missing=3600))
        if storage is not None:
            for user in storage.users:
                schema.add(SchemaNode(String(), name='usr_%d' % user.user_id,
                validator=OneOf(STORAGE_PERMS.keys())))
            for group in storage.groups:
                schema.add(SchemaNode(String(), name='grp_%s' % group.group_id,
                validator=OneOf(STORAGE_PERMS.keys())))

        defaults = {'access': 'open', 'vcs_engine': 'local', 'refresh': 3600}

        return (
            Form(self._request, schema=schema, defaults=defaults, obj=storage),
            TabSet(self._request, STORAGE_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _create(self, values):
        """Create a storage.

        :param values: (dictionary)
            Form values.
        :return: (:class:`~..models.storages.Storage` instance or ``None``)
        """
        # Check existence
        storage_id = values['storage_id'].strip()
        storage = DBSession.query(Storage).filter_by(
            storage_id=storage_id).first()
        if storage is None:
            label = u' '.join(values['label'].strip().split())[0:LABEL_LEN]
            storage = DBSession.query(Storage).filter_by(
                label=label).first()
        if storage is not None:
            self._request.session.flash(
                _('This storage already exists.'), 'alert')
            return

        # Create storage
        # pylint: disable = I0011, W0142
        record = dict(
            [(k, values[k]) for k in values if hasattr(Storage, k)])
        storage = Storage(self._request.registry.settings, **record)
        storage.set_vcs_password(
            self._request.registry.settings, values.get('password'))
        DBSession.add(storage)
        DBSession.commit()

        # Clone
        handler = self._request.registry['handler'].get_handler(
            storage.storage_id, storage)
        handler.clone(self._request)

        return storage

    # -------------------------------------------------------------------------
    @classmethod
    def _save(cls, storage, values):
        """Save a storage settings.

        :param storage: (:class:`~..models.storages.Storage` instance)
            Storage to update.
        :param values: (dictionary)
            Form values.
        :return: (boolean)
            ``True`` if succeeds.
        """
        # Update permissions
        for user in storage.users:
            user.perm = values['perm_%d' % user.user_id]

        DBSession.commit()
        return True

    # -------------------------------------------------------------------------
    def _add_members(self, storage):
        """Add selected users to the storage.

        :param storage_id: (:class:`~..models.storages.Storage` instance)
            Storage object.
        """
        if not has_permission(self._request, 'stg_editor') \
               and self._request.session['storage']['perm'] != 'editor':
            return

        user_ids = [k.user_id for k in storage.users]
        for user_id in get_action(self._request)[1]:
            user_id = int(user_id)
            if not user_id in user_ids:
                storage.users.append(StorageUser(
                    storage.storage_id, user_id))
        DBSession.commit()
        del self._request.session['menu']

    # -------------------------------------------------------------------------
    def _add_member_groups(self, storage):
        """Add selected groups to the storage.

        :param storage_id: (:class:`~..models.storages.Storage` instance)
            Storage object.
        """
        if not has_permission(self._request, 'stg_editor') \
               and self._request.session['storage']['perm'] != 'editor':
            return

        group_ids = [k.group_id for k in storage.groups]
        for group_id in get_action(self._request)[1]:
            if not group_id in group_ids:
                storage.groups.append(StorageGroup(
                    storage.storage_id, group_id))
        DBSession.commit()
        del self._request.session['menu']

    # -------------------------------------------------------------------------
    def _html_path(self, storage_id, path):
        """Return relative path in storage, real path and path in one string
        with ``<a>`` tags.

        :param storage_id: (string)
            Storage ID.
        :param path: (list)
            Splitted relative path inside the storage.
        :return: (tuple)
            A tuple such as ``(path, real_path, html_path)`` or a
            :class:`pyramid.httpexceptions.`HTTPNotFound` exception..
        """
        # Path exists?
        # pylint: disable = I0011, W0142
        real_path = join(self._request.registry.settings['storage.root'],
                         storage_id, *path)
        if not isdir(real_path) and path:
            real_path = dirname(real_path)
            path = path[0:-1]
        if not isdir(real_path):
            raise HTTPNotFound(comment=_('This directory does not exist!'))

        # Root path
        if len(path) == 0:
            return '', real_path, HTML.span(storage_id)

        # Other path
        html = HTML.a(storage_id, href=self._request.route_path(
            'storage_root', storage_id=storage_id))
        for index in range(len(path) - 1):
            html += ' / ' + HTML.a(path[index], href=self._request.route_path(
                'storage_browse', storage_id=storage_id,
                path='/'.join(path[0:index + 1])).decode('utf8'))
        html += ' / ' + HTML.span(path[-1])

        return join('.', *path), real_path, html

    # -------------------------------------------------------------------------
    def _browse_action(self, storage, handler, path):
        """Return form, current action and working status for browse view.

        :param storage: (dictionary)
            Storage dictionary.
        :param handler: (:class:`~..lib.handler.Handler` instance)
            Storage Control System.
        :param path:
            Relative path in storage.
        :return: (tuple)
            A tuple such as ``(form, action, items, working)``
        """
        working = False
        action, items = get_action(self._request)
        form = self._vcs_user_form(storage, action)

        # Check
        if action[0:4] in ('upl!', 'dir!', 'ren!', 'del!'):
            if storage['perm'] == 'user':
                raise HTTPForbidden()
            if not form.validate():
                action = '%s?%s' % (action[0:3], action[4:])
                return form, action, items, working

        # Action
        if action[0:4] == 'upl!':
            upload_file = self._request.params.get('upload_file')
            if not isinstance(upload_file, basestring):
                working = handler.launch(self._request, handler.upload,
                    (vcs_user(self._request, storage), path, upload_file,
                     items and items[0] or None,
                     form.values.get('message', '-')))
        elif action == 'dir!':
            working = handler.launch(
                self._request, handler.mkdir, (path, form.values['directory']))
        elif action[0:4] == 'ren!':
            working = handler.launch(self._request, handler.rename,
                (vcs_user(self._request, storage), path, action[4:],
                 form.values['new_name'], form.values.get('message', '-')))
            action = ''
        elif action[0:4] == 'del!':
            working = handler.launch(self._request, handler.remove,
                (vcs_user(self._request, storage), path, items,
                 form.values.get('message', '-')))
        elif action == 'sch!' and 'search' in self._request.session:
            del self._request.session['search']
        elif action[0:4] in ('npk!', 'cls!', 'pck!', 'prc!'):
            items = self._browse_action_container(
                action, items, join(storage['storage_id'], path), form)

        # Synchronize
        if not working:
            working = handler.synchronize(self._request, action == 'syn!')
            action = '' if working else action

        return form, action, items, working

    # -------------------------------------------------------------------------
    def _browse_action_container(self, action, items, path, form):
        """Manage container actions.

        :param action: (string)
            Action.
        :param items: (list)
            List of selected items.
        :param path: (string)
            Relative path from storage root.
        :param form: (:class:`~..lib.form.Form` instance)
            Browse form.
        :return: (tuple)
            A tuple such as ``(project_id, pack_id)`` or ``items``
        """
        # Close container
        if action == 'cls!' and 'container' in self._request.session:
            del self._request.session['container']

        # Add to container
        elif action[0:4] in ('pck!', 'prc!'):
            add2container(
                self._request, action[0:3], action[4:7], path, items)
            form.forget('#')

        # Create pack
        elif action[0:4] == 'npk!' and 'project' in self._request.session:
            return create_pack(self._request, items, path)

        return items

    # -------------------------------------------------------------------------
    def _vcs_user_form(self, storage, action):
        """Create a VCS user form and, eventually, update ``StorageUser``
        table setting up ``vcs_user`` and ``vcs_password``.

        :param storage: (dictionary)
            Storage dictionary.
        :param action: (string)
            Current action
        :return: (:class:`~.lib.form.Form` instance)
        """
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(String(), name='vcs_user',
            validator=Length(max=ID_LEN), missing=None))
        schema.add(SchemaNode(String(), name='vcs_password', missing=None))
        if 'message' in self._request.params:
            schema.add(SchemaNode(String(), name='message',
                validator=Length(max=DESCRIPTION_LEN)))
        if 'directory' in self._request.params:
            schema.add(SchemaNode(String(), name='directory',
                validator=Length(max=PATH_LEN)))
        elif action[0:4] == 'ren?' or 'new_name' in self._request.params:
            schema.add(SchemaNode(String(), name='new_name',
                validator=Length(max=PATH_LEN)))

        form = Form(self._request, schema=schema)
        if  action[0:4] == 'ren?':
            form.values['new_name'] = action[4:]
            return form
        if not form.validate() or not 'vcs_change' in self._request.params:
            return form

        user = DBSession.query(StorageUser).filter_by(
            storage_id=storage['storage_id'],
            user_id=self._request.session['user_id']).first()
        if user is None:
            user = StorageUser(storage['storage_id'],
                                 user_id=self._request.session['user_id'])
        user.vcs_user = form.values['vcs_user']
        user.vcs_password = encrypt(form.values['vcs_password'],
            self._request.registry.settings['auth.key'])
        DBSession.commit()
        self._request.session.flash(_('VCS user has been changed.'))

        return form



# $Id: renderer.py ada86c3ea608 2012/03/22 22:52:47 patrick $
# -*- coding: utf-8 -*-
"""Renderer functions."""

import webhelpers.paginate as paginate
from webhelpers.html import literal, HTML, tags

from pyramid.i18n import get_localizer

from ..lib.utils import _, has_permission
from ..models import DBSession
from ..models.users import PAGE_SIZE
from ..models.storages import Storage, StorageUser
from ..models.projects import Project, ProjectUser


# =============================================================================
def sortable_column(request, label, sort, current_sorting=None,
                    paging_id=None):
    """Output a header of column with `sort up` and `sort down` buttons.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    :param label: (string)
         Label of column.
    :param sort: (string)
         Sort criteria.
    :param current_sorting: (string, optional)
        Default current sorting.
    :param paging_id: (string, optional)
        Paging ID.
    :return: (literal HTML string)
    """
    img = '/Static/Images/sort_%s.png'
    current = request.params.get('sort') or current_sorting
    query_string = {}
    if request.GET:
        query_string.update(request.GET)
    if paging_id:
        query_string.update({'paging_id': paging_id})

    query_string['sort'] = '+%s' % sort
    url = request.current_route_path(_query=query_string)
    xhtml = '+%s' % sort == current and tags.image(img % 'down_off', 'Down')\
            or HTML.a(tags.image(img % 'down', 'Down'), href=url)

    query_string['sort'] = '-%s' % sort
    url = request.current_route_path(_query=query_string)
    xhtml += '-%s' % sort == current and tags.image(img % 'up_off', 'Up')\
             or HTML.a(tags.image(img % 'up', 'Up'), href=url)

    return literal('%s %s' % (label, xhtml))


# =============================================================================
class Breadcrumbs(object):
    """User breadcrumb trail, current title page and back URL management.

    This class uses session and stores its history in
    ``session['breadcrumbs']``. It is a list of crumbs. Each crumb is a tuple
    such as ``(<title>, <route_name>, <route_parts>)``
    """

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        self._request = request

    # -------------------------------------------------------------------------
    def current_title(self):
        """Title of current page."""
        session = self._request.session
        return ('breadcrumbs' in session and len(session['breadcrumbs']) > 0
                and session['breadcrumbs'][-1][0]) or ''

    # -------------------------------------------------------------------------
    def current_path(self):
        """Path of current page."""
        if not 'breadcrumbs' in self._request.session:
            return self._request.route_path('home')
        crumb = self._request.session['breadcrumbs'][-1]
        return self._request.route_path(crumb[1], **crumb[2])

    # -------------------------------------------------------------------------
    def trail(self):
        """Output XHTML breadcrumb trail."""
        if not 'breadcrumbs' in self._request.session \
               or len(self._request.session['breadcrumbs']) < 2:
            return literal('&nbsp;')

        translate = get_localizer(self._request).translate
        crumbs = []
        for crumb in self._request.session['breadcrumbs'][0:-1]:
            if crumb[1] is not None:
                crumbs.append(u'<a href="%s">%s</a>' % (
                    self._request.route_path(crumb[1], **crumb[2]),
                    translate(crumb[0])))
            else:
                crumbs.append(translate(crumb[0]))
        return literal(u' » '.join(crumbs))

    # -------------------------------------------------------------------------
    def back_title(self):
        """Output title of previous page."""
        if not 'breadcrumbs' in self._request.session \
               or len(self._request.session['breadcrumbs']) < 2:
            return _('home')
        return get_localizer(self._request).translate(
            self._request.session['breadcrumbs'][-2][0])

    # -------------------------------------------------------------------------
    def back_button(self):
        """A button to return to the previous page."""
        return literal(
            u'<a href="{url}" title="{title}">'
            '<img src="/Static/Images/back.png" alt="Back"/></a>'.format(
                url=self.back_path(), title=self.back_title()))

    # -------------------------------------------------------------------------
    def back_path(self):
        """Output the path of previous page."""
        if not 'breadcrumbs' in self._request.session \
               or len(self._request.session['breadcrumbs']) < 2:
            return self._request.route_path('home')
        crumb = self._request.session['breadcrumbs'][-2]
        return self._request.route_path(crumb[1], **crumb[2])

    # -------------------------------------------------------------------------
    def add(self, title, length=6, root_chunks=6, replace=None):
        """Add a crumb in breadcrumb trail.

        :param title: (string)
            Page title in breadcrumb trail.
        :param length: (int, default=6)
            Maximum crumb number. If 0, it keeps the current length.
        :param root_chunks: (int, default=6)
            Number of path chunks to compare to highlight menu item.
        :param replace: (string, optional):
            If current path is ``replace``, this method call :meth:`pop` before
            any action.
        """
        # Environment
        session = self._request.session
        if not 'breadcrumbs' in session:
            session['breadcrumbs'] = [(_('Home'), 'home', {}, 1)]
        if not length:
            length = len(session['breadcrumbs'])

        # Replace
        if replace and self.current_path() == replace:
            self.pop()

        # Scan old breadcrumb trail to find the right position
        route_name = (self._request.matched_route
                      and self._request.matched_route.name)
        if route_name is None:
            return
        compare_name = route_name.replace('_root', '_browse')
        crumbs = []
        for crumb in session['breadcrumbs']:
            if len(crumbs) >= length - 1 or not crumb[1] \
                   or crumb[1].replace('_root', '_browse') == compare_name:
                break
            crumbs.append(crumb)

        # Add new breadcrumb
        crumbs.append((
            title, route_name, self._request.matchdict, root_chunks))
        session['breadcrumbs'] = crumbs

    # -------------------------------------------------------------------------
    def pop(self):
        """Pop last breadcrumb."""
        session = self._request.session
        if 'breadcrumbs' in session and len(session['breadcrumbs']) > 1:
            session['breadcrumbs'] = session['breadcrumbs'][0:-1]


# =============================================================================
class Menu(object):
    """User menu management."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        self._request = request
        self._translate = get_localizer(request).translate

    # -------------------------------------------------------------------------
    def xhtml(self):
        """Output XHTML user menu."""
        if not 'user_id' in self._request.session:
            return ''
        if 'menu' in self._request.session:
            return self._make_xhtml()
        menu = []

        # Storages
        if has_permission(self._request, 'stg_user'):
            submenu = [self._entry(_('Advanced search'), 'file_search'),
                       self._entry(_('All storages'), 'storage_index')]
            self._storage_entries(self._request.session['user_id'], submenu)
            menu.append(self._entry(_('Storages'), None, submenu))

        # Projects
        if has_permission(self._request, 'prj_user'):
            submenu = [self._entry(_('All projects'), 'project_index')]
            self._project_entries(self._request.session['user_id'], submenu)
            menu.append(self._entry(_('Projects'), None, submenu))

        # Administration
        submenu = []
        if has_permission(self._request, 'admin'):
            submenu.append(self._entry(_('Site'), 'site_admin'))
        if has_permission(self._request, 'usr_editor'):
            submenu.append(self._entry(_('Users'), 'user_admin'))
        if has_permission(self._request, 'grp_editor'):
            submenu.append(self._entry(_('Groups'), 'group_admin'))
        if has_permission(self._request, 'stg_editor'):
            submenu.append(self._entry(_('Storages'), 'storage_admin'))
        if has_permission(self._request, 'prj_editor'):
            submenu.append(self._entry(_('Projects'), 'project_admin'))
        if submenu:
            menu.append(self._entry(_('Administration'), None, submenu))

        if not menu:
            return ''
        self._request.session['menu'] = tuple(menu)
        return self._make_xhtml()

    # -------------------------------------------------------------------------
    def _make_xhtml(self):
        """Return an <ul> structure with current highlighted current entry."""
        # Make XHTML
        menu = self._request.session['menu']
        xhtml = '<ul>%s</ul>' % self._make_xhtml_entries(menu, 0)

        # Highlight current entry
        if 'breadcrumbs' in self._request.session:
            for crumb in reversed(self._request.session['breadcrumbs'][1:]):
                if crumb[1] is None:
                    continue
                path = self._request.route_path(crumb[1], **crumb[2])
                path = '/'.join(path.split('/')[0:crumb[3] + 1])\
                       .replace('/edit/', '/view/')
                if 'href="%s"' % path in xhtml:
                    xhtml = xhtml.replace('<a class="slow" href="%s"' % path,
                        '<a class="slow current" href="%s"' % path)
                    break

        # Tag current project
        if 'project' in  self._request.session:
            path = self._request.route_path('project_dashboard',
                project_id=self._request.session['project']['project_id'],
                _query={'id': self._request.session['project']['project_id']})
            xhtml = xhtml.replace('<li><a class="slow" href="%s"' % path,
                '<li class="active"><a class="slow" href="%s"' % path)

        return literal(xhtml)

    # -------------------------------------------------------------------------
    def _make_xhtml_entries(self, entries, depth):
        """Return <li> tags with entries.

        :param entries: (tuple)
            Tuple of entry tuples (See :meth:`_entry`)
        :param depth: (integer)
            Depth of entries in menu.
        """
        xhtml = ''
        for entry in entries:
            tag = (depth == 0 and '<strong>') \
                  or (depth == 1 and entry[2] and '<em>') or ''
            xhtml += '<li>' \
                + (entry[1] and '<a class="slow" href="%s">' % entry[1] or '')\
                + tag + entry[0] + tag.replace('<', '</') \
                + (entry[1] and '</a>' or '')
            if entry[3]:
                xhtml += '<ul>%s</ul>' % self._make_xhtml_entries(
                    entry[3], depth + 1)
            xhtml += '</li>'
        return xhtml

    # -------------------------------------------------------------------------
    def _storage_entries(self, user_id, submenu):
        """Update menu entries for user storages shown in menu.

        :param user_id: (string)
            Current user ID.
        :param submenu: (list)
            Current submenu list.
        """
        # Look for user storages
        for storage in DBSession.query(Storage).join(StorageUser)\
                .filter(StorageUser.user_id == user_id)\
                .filter(StorageUser.in_menu == True).order_by(Storage.label):
            submenu.append(self._entry(
                storage.label, 'storage_root',
                None, True, storage_id=storage.storage_id))

    # -------------------------------------------------------------------------
    def _project_entries(self, user_id, submenu):
        """Update menu entries for user projects shown in menu.

        :param user_id: (string)
            Current user ID.
        :param submenu: (list)
            Current submenu list.
        """
        # Look for user projects
        for project in DBSession\
                .query(Project.project_id, Project.label, ProjectUser.perm)\
                .join(ProjectUser).filter(ProjectUser.user_id == user_id)\
                .filter(ProjectUser.in_menu == True).order_by(Project.label):
            pid = project[0]
            subentries = [
                self._entry(_('Dashboard'), 'project_dashboard',
                            None, True, project_id=pid),
                self._entry(_('Last results'), 'project_results',
                            None, True, project_id=pid)]
            if project[2] == 'editor' \
                   or has_permission(self._request, 'prj_editor'):
                subentries.append(self._entry(_('Settings'), 'project_view',
                    None, True, project_id=pid))
            submenu.append(self._entry(project[1], 'project_dashboard',
                tuple(subentries), True, project_id=pid, _query={'id': pid}))

    # -------------------------------------------------------------------------
    def _entry(self, label, route_name, subentries=None, is_minor=False,
               **kwargs):
        """A menu entry tuple.

        :param label: (string)
            Label of the entry.
        :param route_name: (string)
            Name of the route for the link.
        :param subentries: (list, optional)
            List of subentries.
        :param is_minor: (boolean, default=False)
            Indicate whether this entry is a minor one.
        :param kwargs: (dictionary)
            Keyworded arguments for :meth:`pyramid.request.Request.route_path`.
        :return: (tuple)
            A tuple such as
            ``(label, url, is_minor, (subentry, subentry...))``.
        """
        return (self._translate(label),
                route_name and self._request.route_path(route_name, **kwargs),
                is_minor, subentries and tuple(subentries))


# =============================================================================
class Paging(paginate.Page):
    """An extended :class:`webhelpers.paginate.Page` that manages one page of a
    splitted list or results from ORM queries.

    This class uses three parameters in request: ``page_size`` and ``page``.

    It stores its information and filters definitions in ``session['paging']``.
    This structure looks like:

    ``session['paging'] = (default_page_size, {'users': {'size': 20, 'page': 3,
    'sort': 'name', 'f_id': 'smith', 'f_group': 'managers',...},...})``
    """

    page_sizes = ('', 5, 10, 20, 40, 80, 160, 320)

    # -------------------------------------------------------------------------
    def __init__(self, request, paging_id, collection, **kwargs):
        """Constructor method."""
        self.paging_id = paging_id
        self._request = request

        # Update paging session
        params = request.session['paging'][1][paging_id]
        if request.params.get('page_size'):
            params['size'] = int(request.params['page_size'])
        if request.params.get('page'):
            params['page'] = int(request.params['page'])

        # Return page
        get_params = self._request.GET.copy()
        get_params.update({'paging_id': paging_id})
        super(Paging, self).__init__(
            collection, params['page'], params['size'],
            url=paginate.PageURL(self._request.path, get_params), **kwargs)

    # -------------------------------------------------------------------------
    @classmethod
    def params(cls, request, paging_id, default_sorting):
        """Return current paging parameters: filter criteria and sorting.

        :param paging_id: (string)
            Paging ID.
        :param default_sorting: (string)
            Default sorting.
        :return: (dictionary)
            The paging dictionary. See :class:`~.renderer.Paging` class.
        """
        if not 'paging' in request.session:
            request.session['paging'] = (PAGE_SIZE, {})
        if not paging_id in request.session['paging'][1]:
            request.session['paging'][1][paging_id] = {
                'size': request.session['paging'][0],
                'page': 1, 'sort': default_sorting}
        if len(request.POST):
            params = request.session['paging'][1][paging_id]
            request.session['paging'][1][paging_id] = {
                'size': params['size'], 'page': params['page'],
                'sort': params['sort']}

        params = request.session['paging'][1][paging_id]
        for key in request.params:
            if (key[0:2] == 'f_' or key == 'sort') and request.params[key]:
                params[key] = request.params[key]

        return params

    # -------------------------------------------------------------------------
    def pager_top(self, icon_name=None, label=None):
        """Output a string with links to first, previous, next and last pages.

        :param icon_name: (string, optional)
            Name of the icon representing the items.
        :param label: (string, optional)
            Label representing the items.
        :return: (string)
        """
        icon = ''
        if icon_name is not None and label is not None:
            icon = literal('<img src="/Static/Images/{i}.png" alt="{l}"'
                        ' title="{l}"/>  &nbsp;'.format(i=icon_name, l=label))
        img = '<img src="/Static/Images/go_%s.png" alt="%s"/>'
        return icon + self.pager('$link_first $link_previous '
            '<span>$first_item &ndash; $last_item</span> / $item_count '
            '$link_next $link_last',
            symbol_first=literal(img % ('first', 'First')),
            symbol_previous=literal(img % ('previous', 'Previous')),
            symbol_next=literal(img % ('next', 'Next')),
            symbol_last=literal(img % ('last', 'Last'))) or literal('&nbsp;')

    # -------------------------------------------------------------------------
    def pager_bottom(self, icon_name=None, label=None):
        """Output a string with links to some previous and next pages.

        :param icon_name: (string, optional)
            URL of the icon representing the items.
        :param label: (string, optional)
            Label representing the items.
        :return: (string)
        """
        icon = ''
        if icon_name is not None and label is not None:
            icon = literal('<img src="/Static/Images/{i}.png" alt="{l}"'
                       ' title="{l}"/> &nbsp;'.format(i=icon_name, l=label))
        return (icon + self.pager('~3~')) or literal('&nbsp;')

    # -------------------------------------------------------------------------
    def sortable_column(self, label, sort):
        """Output a header of column with `sort up` and `sort down` buttons.

        See :func:`~.lib.utils.sortable_column`.

        :param label: (string)
             Label of column.
        :param sort: (string)
             Sort criteria.
        :return: (literal HTML string)
        """
        return sortable_column(self._request, label, sort,
            self._request.session['paging'][1][self.paging_id]['sort'],
            self.paging_id)

    # -------------------------------------------------------------------------
    @classmethod
    def pipe(cls):
        """Output a pipe image."""
        return literal(
            '<img src="/Static/Images/action_pipe.png" alt="Pipe"/>')


# =============================================================================
class TabSet(object):
    """A class to manages tabs."""

    # -------------------------------------------------------------------------
    def __init__(self, request, labels):
        """Constructor method."""
        self._request = request
        self.labels = labels

    # -------------------------------------------------------------------------
    def toc(self, tab_id):
        """Output a table of content of the ``TabSet`` in an ``<ul>``
        structure.

        :param tab_id: (string)
            Tab set ID.
        :return: (string)
            ``<ul>`` structure.
        """
        translate = get_localizer(self._request).translate
        xml = '<ul id="%s" class="tabToc">\n' % tab_id
        for index, label in enumerate(self.labels):
            xml += '  <li><a class="tabLink" id="tabLink%d" href="#tab%d">' \
                   '<span>%s</span></a></li>\n' % (
                index, index, translate(label))
        xml += '</ul>\n'
        return literal(xml)

    # -------------------------------------------------------------------------
    def tab_begin(self, index, access_key=None):
        """Open a tab zone.

        :param index: (integer)
            Tab index.
        :param access_key: (string, optional)
            Access key for tab.
        :return: (string)
            Opening ``fieldset`` structure with legend.
        """
        return literal('<fieldset class="tabContent" id="tab%d">\n'
                       '  <legend%s><span>%s</span></legend>\n' % (
            index, access_key and ' accesskey="%s"' % access_key or '',
            self.labels[index]))

    # -------------------------------------------------------------------------
    @classmethod
    def tab_end(cls):
        """Close a tab zone."""
        return literal('</fieldset>')



# $Id: storages.py 86b088d85a5a 2012/03/21 22:20:47 patrick $
# -*- coding: utf-8 -*-
"""SQLAlchemy-powered model definition for storages."""

from os.path import exists, join
import shutil
from lxml import etree
from textwrap import fill
from sqlalchemy import Column, ForeignKey, types
from sqlalchemy.orm import relationship

from ..lib.utils import _, encrypt, normalize_name
from . import ID_LEN, LABEL_LEN, DESCRIPTION_LEN, PATH_LEN, Base, DBSession
from .users import User
from .groups import Group


STORAGE_ACCESS = {'open': _('open'), 'restricted': _('restricted'),
                    'closed': _('closed')}
VCS_ENGINES = {'none': _(u'none – None'), 'local': _(u'local – Local'),
               'hg': u'hg – Mercurial', 'svn': u'svn – Subversion',
               'git': u'git – Git'}
STORAGE_PERMS = {'editor': _('File editor'), 'user': _('User')}
REFRESH = 3600
XML_NS = '{http://www.w3.org/XML/1998/namespace}'


# =============================================================================
class Storage(Base):
    """SQLAlchemy-powered storage model."""
    # pylint: disable = I0011, W0142, R0902

    __tablename__ = 'storages'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    storage_id = Column(types.String(ID_LEN), primary_key=True)
    label = Column(types.String(LABEL_LEN), unique=True, nullable=False)
    description = Column(types.String(DESCRIPTION_LEN))
    vcs_engine = Column(
        types.Enum(*VCS_ENGINES.keys(), name='vcs_engine_enum'),
        nullable=False)
    vcs_url = Column(types.String(PATH_LEN))
    vcs_user = Column(types.String(ID_LEN))
    vcs_password = Column(types.String(40))
    public_url = Column(types.String(PATH_LEN))
    access = Column(
        types.Enum(*STORAGE_ACCESS.keys(), name='stg_access_enum'),
        nullable=False, default='open')
    refresh = Column(types.Integer, default=REFRESH)
    users = relationship('StorageUser', backref='storage')
    groups = relationship('StorageGroup')

    # -------------------------------------------------------------------------
    def __init__(self, settings, storage_id, label, description=None,
                 vcs_engine='local', vcs_url=None, vcs_user=None,
                 vcs_password=None, public_url=None, access=None,
                 refresh=None):
        """Constructor method."""
        # pylint: disable = I0011, R0913
        self.storage_id = storage_id.strip()[0:ID_LEN]
        self.label = u' '.join(label.strip().split())[0:LABEL_LEN]
        if description:
            self.description = \
                u' '.join(description.strip().split())[0:DESCRIPTION_LEN]
        self.access = access
        self.vcs_engine = vcs_engine
        if not vcs_engine in ('none', 'local'):
            self.vcs_url = vcs_url[0:PATH_LEN]
            self.vcs_user = vcs_user and vcs_user.strip()[0:ID_LEN]
            self.set_vcs_password(settings, vcs_password)
        elif public_url is not None:
            self.public_url = public_url.strip()[0:PATH_LEN]
        self.refresh = refresh

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return u"<Storage('{storage_id}', '{label}', '{access}', "\
            "'{vcs_engine}', '{vcs_url}', '{vcs_user}', {refresh})>".format(
            storage_id=self.storage_id, label=self.label,
            access=self.access, vcs_engine=self.vcs_engine,
            vcs_url=self.vcs_url, vcs_user=self.vcs_user,
            refresh=self.refresh or 0).encode('utf8')

    # -------------------------------------------------------------------------
    def set_vcs_password(self, settings, vcs_password):
        """Encrypt and set password.

        :param settings: (dictionary)
            Pyramid deployment settings.
        :param vcs_password: (string)
            Clear VCS password.
        """
        if vcs_password:
            self.vcs_password = encrypt(
                vcs_password.strip(), settings['auth.key'])

    # -------------------------------------------------------------------------
    @classmethod
    def delete(cls, storage_root, storage_id):
        """Delete a storage.

        :param storage_root: (string)
            Storage root path.
        :param storage_id: (string)
            Storage identifier.
        """
        # Delete records in database
        DBSession.query(cls).filter_by(storage_id=storage_id).delete()
        DBSession.commit()

        # Remove directory
        if exists(join(storage_root, storage_id)):
            shutil.rmtree(join(storage_root, storage_id))

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, settings, storage_elt, error_if_exists=True):
        """Load a storage from a XML file.

        :param settings: (dictionary)
            Application settings.
        :param storage_elt: (:class:`lxml.etree.Element` instance)
            Storage XML element.
        :param error_if_exists: (boolean, default=True)
            It returns an error if storage already exists.
        :return: (:class:`pyramid.i18n.TranslationString` ``None`` or
            :class:`Storage` instance)
            Error message or ``None`` or the new storage object.
        """
        # Reset
        storage_id = storage_elt.get('%sid' % XML_NS).strip()[0:ID_LEN]
        label = u' '.join(
            storage_elt.findtext('label').strip().split())[0:LABEL_LEN]
        if storage_elt.find('reset') is not None \
               and bool(storage_elt.findtext('reset')):
            Storage.delete(settings['storage.root'], storage_id)

        # Check if already exists
        storage = DBSession.query(cls).filter_by(
            storage_id=storage_id).first()
        if storage is None:
            storage = DBSession.query(cls).filter_by(label=label).first()
        if storage is not None:
            if error_if_exists:
                return _(
                    'Storage "${i}" already exists.', {'i': storage_id})
            return

        # Create storage
        # pylint: disable = I0011, W0142
        vcs_elt = storage_elt.find('vcs')
        record = {
            'label': label,
            'description': storage_elt.findtext('description'),
            'vcs_engine': vcs_elt.get('engine'),
            'vcs_url': vcs_elt.findtext('url') is not None
                and vcs_elt.findtext('url').strip() or None,
            'vcs_user': vcs_elt.findtext('user') is not None
                and vcs_elt.findtext('user').strip() or None,
            'vcs_password': vcs_elt.findtext('password') is not None
                and vcs_elt.findtext('password').strip() or None,
            'public_url': vcs_elt.findtext('public') is not None
                and vcs_elt.findtext('public').strip() or None,
            'access': storage_elt.findtext('access') is not None
                and storage_elt.findtext('access').strip() or 'open',
            'refresh': storage_elt.findtext('refresh') is not None
                and int(storage_elt.findtext('refresh').strip()) or None}
        storage = Storage(settings, storage_id, **record)
        DBSession.add(storage)
        DBSession.commit()

        # Add users
        done = []
        for item in storage_elt.findall('members/member'):
            login = normalize_name(item.text)[0:ID_LEN]
            if login not in done:
                user = DBSession.query(User).filter_by(login=login).first()
                if user is not None:
                    storage.users.append(StorageUser(
                        storage.storage_id, user.user_id,
                        item.get('in-menu'), item.get('permission', 'user')))
                done.append(login)

        # Add groups
        done = []
        for item in storage_elt.findall('members/member-group'):
            group_id = normalize_name(item.text)[0:ID_LEN]
            if group_id not in done:
                group = DBSession.query(Group).filter_by(
                    group_id=group_id).first()
                if group is not None:
                    storage.groups.append(StorageGroup(storage.storage_id,
                        group.group_id, item.get('permission', 'user')))
                done.append(group_id)

        DBSession.commit()
        return storage

    # -------------------------------------------------------------------------
    def xml(self):
        """Serialize a storage to a XML representation.

        :return: (:class:`lxml.etree.Element`)
        """
        storage_elt = etree.Element('storage')
        storage_elt.set('%sid' % XML_NS, self.storage_id)
        etree.SubElement(storage_elt, 'label').text = self.label
        if self.description:
            etree.SubElement(storage_elt, 'description').text = fill(
                self.description, initial_indent='\n' + ' ' * 8,
                subsequent_indent=' ' * 8)
        elt = etree.SubElement(storage_elt, 'vcs')
        elt.set('engine', self.vcs_engine)
        if self.vcs_url:
            etree.SubElement(elt, 'url').text = self.vcs_url
        if self.vcs_user:
            etree.SubElement(elt, 'user').text = self.vcs_user
        if self.public_url:
            etree.SubElement(elt, 'public').text = self.public_url
        if self.access != 'open':
            etree.SubElement(storage_elt, 'access').text = self.access
        if self.refresh != REFRESH:
            etree.SubElement(storage_elt, 'refresh').text = str(self.refresh)

        # Members
        if self.users or self.groups:
            members_elt = etree.SubElement(storage_elt, 'members')
            for user in self.users:
                elt = etree.SubElement(members_elt, 'member')
                elt.text = user.user.login
                if user.in_menu:
                    elt.set('in-menu', 'true')
                if user.perm != 'user':
                    elt.set('permission', user.perm or 'none')
            for group in self.groups:
                elt = etree.SubElement(members_elt, 'member-group')
                elt.text = group.group_id
                if group.perm != 'user':
                    elt.set('permission', group.perm)

        return storage_elt


# =============================================================================
class StorageUser(Base):
    """SQLAlchemy-powered association table between ``Storage`` and
    ``User``."""
    # pylint: disable = I0011, R0913, W0142

    __tablename__ = 'storages_users'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    storage_id = Column(types.String(ID_LEN), ForeignKey(
        'storages.storage_id', onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True)
    user_id = Column(types.Integer, ForeignKey('users.user_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    in_menu = Column(types.Boolean(), default=False)
    perm = Column(types.Enum(*STORAGE_PERMS.keys(), name='stg_perms_enum'))
    vcs_user = Column(types.String(ID_LEN))
    vcs_password = Column(types.String(40))
    user = relationship('User')

    # -------------------------------------------------------------------------
    def __init__(self, storage_id, user_id, in_menu=False, perm=None,
                 vcs_user=None, vcs_password=None, settings=None):
        """Constructor method."""
        self.storage_id = storage_id.strip()[0:ID_LEN]
        self.user_id = user_id
        self.in_menu = bool(in_menu)
        if perm != 'none':
            self.perm = perm
        if vcs_user and settings is not None:
            self.vcs_user = vcs_user
            self.set_vcs_password(settings, vcs_password)

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<StorageUser('{storage_id}', {user_id}, {in_menu}, "\
               "'{perm}', '{vcs_user}')>".format(
            storage_id=self.storage_id, user_id=self.user_id,
            in_menu=self.in_menu, perm=self.perm, vcs_user=self.vcs_user or '')

    # -------------------------------------------------------------------------
    def set_vcs_password(self, settings, vcs_password):
        """Encrypt and set VCS password.

        :param settings: (dictionary)
            Pyramid deployment settings.
        :param password: (string)
            Clear password.
        """
        self.vcs_password = encrypt(vcs_password, settings['auth.key'])


# =============================================================================
class StorageGroup(Base):
    """SQLAlchemy-powered association table between ``Storage`` and
    ``Group``."""
    # pylint: disable = I0011, W0142, R0903

    __tablename__ = 'storages_groups'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    storage_id = Column(types.String(ID_LEN), ForeignKey(
        'storages.storage_id', onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True)
    group_id = Column(types.String(ID_LEN), ForeignKey('groups.group_id',
        onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    perm = Column(types.Enum(*STORAGE_PERMS.keys(), name='stg_perms_enum'),
                  default='user')

    # -------------------------------------------------------------------------
    def __init__(self, storage_id, group_id, perm=None):
        """Constructor method."""
        self.storage_id = storage_id.strip()[0:ID_LEN]
        self.group_id = group_id.strip()[0:ID_LEN]
        self.perm = perm

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<StorageGroup('{storage_id}', '{group_id}', '{perm}')>".format(
            storage_id=self.storage_id, group_id=self.group_id, perm=self.perm)



# $Id: publiset.py ada86c3ea608 2012/03/22 22:52:47 patrick $
"""Publiset management."""

from os.path import join, basename, splitext
from lxml import etree
from copy import deepcopy
from cStringIO import StringIO

from ...xml import XML_NS, load

PUBLIDOC_VERSION = '1.0'
PUBLISET_VERSION = '1.0'


# =============================================================================
class Publiset(object):
    """Class for Publiset management."""

    # -------------------------------------------------------------------------
    def __init__(self, engine, base_path):
        """Constructor method.

        :param engine: (:class:`~.lib.engine.leprisme.Engine` instance)
            Engine object on which it depends.
        :param base_path: (string)
            Base path for files in publiset.
        """
        self._engine = engine
        self._base_path = base_path

    # -------------------------------------------------------------------------
    def fullname(self, file_elt):
        """Find the full path of a file from a file tag.

        :pgaram file_elt: (etree.Element object)
             File element.
        :return: (string)
             Full path name.
        """
        elt = file_elt.getparent()
        while elt is not None and elt.get('path') is None:
            elt = elt.getparent()

        if elt is None:
            return join(self._base_path, file_elt.get('name'))
        return join(self._base_path, elt.get('path'), file_elt.get('name'))

    # -------------------------------------------------------------------------
    def compose(self, filename, set_root):
        """Compose an XML document from a publiset XML composition.

        :param filename: (string)
            Name of the composition file.
        :param set_root: (:class:`lxml.etree.Element` instance)
            ``composition`` element.
        :return: (etree.ElementTree object)
            Document tree or ``None``.
        """
        # Document root element
        doc_root = self._doc_element(
            set_root, 'publidoc', {'version': PUBLIDOC_VERSION})

        # Browse structure
        for elt in set_root.iterchildren(tag=etree.Element):
            if not self._append(elt, doc_root):
                return
        doc_root = etree.ElementTree(doc_root)

        if self._engine.config('Input', 'validate') == 'true':
            error = load(filename, self._engine.relaxngs, doc_root)
            if isinstance(error, basestring):
                self._engine.build.stopped(error, 'a_error')

        return doc_root

    # -------------------------------------------------------------------------
    @classmethod
    def create(cls, element):
        """Create an empty ``publiset`` document and fill it with ``element``.

        :param element: (:class:`lxml.etree.Element` instance)
            Main element.
        :return: (etree.ElementTree)
        """
        root = etree.Element('publiset', version=PUBLISET_VERSION)
        root.append(element)
        return etree.ElementTree(root)

    # -------------------------------------------------------------------------
    def _append(self, set_current, doc_current):
        """Append ``set_current`` to ``doc_current`` element, converting tag
        and attribute names.

        :param set_current: (:class:`lxml.etree.Element` instance)
            Publiset element.
        :param doc_current: (:class:`lxml.etree.Element` instance)
            Target document element.
        :return: (boolean)
        """
        if set_current.tag == 'file':
            return self._append_file(set_current, doc_current)

        doc_elt = self._doc_element(set_current)
        doc_current.append(doc_elt)
        for set_elt in set_current.iterchildren(tag=etree.Element):
            if not self._append(set_elt, doc_elt):
                return False
        return True

    # -------------------------------------------------------------------------
    def _append_file(self, file_elt, doc_current):
        """Append file content as child of ``doc_current``.

        :param file_elt: (:class:`lxml.etree.Element` instance)
             File element.
        :param doc_current: (:class:`lxml.etree.Element` instance)
            Target document element.
        :return: (boolean)
        """
        # Load file
        # pylint: disable = I0011, E1103
        relaxngs = self._engine.relaxngs \
            if self._engine.config('Input', 'validate') == 'true' else None
        filename = self.fullname(file_elt)
        tree = load(filename, relaxngs)
        if isinstance(tree, basestring):
            self._engine.build.stopped(tree, 'a_error')
            return False

        # Find XSL transformation
        set_elt = file_elt
        while set_elt is not None and set_elt.get('xslt') is None:
            set_elt = set_elt.getparent()
        if set_elt is not None:
            fid = '"%s"' % splitext(basename(filename))[0]
            filename = join(self._engine.build.path, 'Processor', 'Xsl',
                            set_elt.get('xslt'))
            try:
                xslt = etree.XSLT(etree.parse(filename))
            except (IOError, etree.XSLTParseError, etree.XMLSyntaxError), err:
                self._engine.build.stopped(
                    str(err).replace(self._engine.build.path, '..'))
                return False
            try:
                tree = etree.ElementTree(xslt(tree, fid=fid).getroot())
            except etree.XSLTApplyError, err:
                self._engine.build.stopped(err, 'a_error')
                return False
            tree = etree.parse(
                StringIO(etree.tostring(tree, pretty_print=True)))

        # Find XPath
        set_elt = file_elt
        while set_elt is not None and set_elt.get('xpath') is None:
            set_elt = set_elt.getparent()

        # Add elements
        if set_elt is None:
            for child in tree.getroot().iterchildren(tag=etree.Element):
                doc_current.append(deepcopy(child))
        else:
            for child in tree.xpath(set_elt.get('xpath')):
                doc_current.append(deepcopy(child))

        return True

    # -------------------------------------------------------------------------
    @classmethod
    def _doc_element(cls, set_elt, default_tag=None, default_atts=None):
        """Create a document element from a set element according to its ``as``
        and ``attributes`` attributes.

        :param set_elt: (:class:`lxml.etree.Element` instance)
            Source element
        :param default_tag: (string)
            If ``default_tag`` is None and ``as`` attribute doesn't exist, the
            ``set_elt`` name is used as tag name.
        :param default_atts: (dictionary)
            If ``default_atts`` is None and ``attributes`` attribute doesn't
            exist, ``set_elt`` attributes are copied.
        :return: (ElementTree)
        """
        # Tag
        if default_tag is None:
            default_tag = set_elt.tag
        doc_elt = etree.Element(set_elt.get('as', default_tag))

        # Attributes
        atts = set_elt.get('attributes', '').split()
        if atts:
            atts = dict([(i.split('=')[0], i.split('=')[1]) for i in atts])
        else:
            atts = default_atts or set_elt.attrib
        for att, value in atts.items():
            if att not in ('as', 'attributes', 'path', 'xslt', 'xpath'):
                doc_elt.set(att.replace('xml:', XML_NS), value)

        # Text
        doc_elt.text = set_elt.text
        doc_elt.tail = set_elt.tail

        return doc_elt



# $Id: groups.py 91a251074d63 2012/03/18 23:07:53 patrick $
"""SQLAlchemy-powered model definition for groups."""

from lxml import etree
from textwrap import fill
from sqlalchemy import Table, Column, ForeignKey, types
from sqlalchemy.orm import relationship

from ..lib.utils import _, normalize_name
from . import Base, DBSession, ID_LEN, LABEL_LEN, DESCRIPTION_LEN
from .users import PERM_SCOPES, USER_PERMS, User


GROUPS_USERS = Table('groups_users', Base.metadata,
    Column('group_id', types.String(ID_LEN), ForeignKey('groups.group_id',
            onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
    Column('user_id', types.Integer,  ForeignKey('users.user_id',
            onupdate='CASCADE', ondelete='CASCADE'), primary_key=True),
    mysql_engine='InnoDB'
    )
XML_NS = '{http://www.w3.org/XML/1998/namespace}'


# =============================================================================
class Group(Base):
    """SQLAlchemy-powered group model."""
    # pylint: disable = I0011, R0903

    __tablename__ = 'groups'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    group_id = Column(types.String(ID_LEN), primary_key=True)
    label = Column(types.String(LABEL_LEN), unique=True, nullable=False)
    description = Column(types.String(DESCRIPTION_LEN))
    perms = relationship('GroupPerm')
    users = relationship('User', secondary=GROUPS_USERS, backref='groups')

    # -------------------------------------------------------------------------
    def __init__(self, group_id, label, description=None):
        """Constructor method."""
        self.group_id = group_id.strip()[0:ID_LEN]
        self.label = u' '.join(label.strip().split())[0:LABEL_LEN]
        if description:
            self.description = \
                u' '.join(description.strip().split())[0:DESCRIPTION_LEN]

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return u"<Group('{group_id}', '{label}')>".format(
            group_id=self.group_id, label=self.label).encode('utf8')

    # -------------------------------------------------------------------------
    @classmethod
    def load(cls, group_elt, error_if_exists=True):
        """Load a user from a XML file.

        :param group_elt: (:class:`lxml.etree.Element` instance)
            Group XML element.
        :param error_if_exists: (boolean, default=True)
            It returns an error if group already exists.
        :return: (:class:`pyramid.i18n.TranslationString` or ``None``)
            Error message or ``None``.
        """
        # Check if already exists
        group_id = group_elt.get('%sid' % XML_NS).strip()[0:ID_LEN]
        label = u' '.join(
            group_elt.findtext('label').split()).strip()[0:LABEL_LEN]
        group = DBSession.query(cls).filter_by(group_id=group_id).first()
        if group is None:
            group = DBSession.query(cls).filter_by(label=label).first()
        if group is not None:
            if error_if_exists:
                return _('Group "${l} (${i})" already exists.',
                         {'i': group_id, 'l': label})
            return

        # Create group
        group = cls(group_id, label, group_elt.findtext('description'))
        DBSession.add(group)
        DBSession.commit()

        # Add permissions
        done = []
        for item in group_elt.findall('permissions/permission'):
            if item.get('scope') not in done:
                group.perms.append(GroupPerm(
                    group.group_id, item.get('scope'), item.text))
                done.append(item.get('scope'))

        # Add members
        done = []
        for item in group_elt.findall('members/member'):
            login = normalize_name(item.text)[0:ID_LEN]
            if login not in done:
                user = DBSession.query(User).filter_by(login=login).first()
                if user is not None:
                    group.users.append(user)
                done.append(login)

        DBSession.commit()

    # -------------------------------------------------------------------------
    def xml(self):
        """Serialize a group to a XML representation.

        :return: (:class:`lxml.etree.Element`)
        """
        group_elt = etree.Element('group')
        group_elt.set('%sid' % XML_NS, self.group_id)
        etree.SubElement(group_elt, 'label').text = self.label
        if self.description:
            etree.SubElement(group_elt, 'description').text = fill(
                self.description, initial_indent='\n' + ' ' * 8,
                subsequent_indent=' ' * 8)

        # Permissions
        if self.perms:
            elt = etree.SubElement(group_elt, 'permissions')
            for perm in self.perms:
                etree.SubElement(
                    elt, 'permission', scope=perm.scope).text = perm.level

        # Members
        if self.users:
            members_elt = etree.SubElement(group_elt, 'members')
            for user in self.users:
                user = DBSession.query(User.login).filter_by(
                    user_id=user.user_id).first()
                elt = etree.SubElement(members_elt, 'member').text = user[0]

        return group_elt


# =============================================================================
class GroupPerm(Base):
    """SQLAlchemy-powered group permission class."""
    # pylint: disable = I0011, R0903, W0142

    __tablename__ = 'groups_perms'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    group_id = Column(types.String(ID_LEN),
        ForeignKey('groups.group_id', onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True)
    scope = Column(types.Enum(*PERM_SCOPES.keys(), name='perm_scope_enum'),
        primary_key=True)
    level = Column(types.Enum(*USER_PERMS.keys(), name='grp_perms_enum'))

    # -------------------------------------------------------------------------
    def __init__(self, group_id, scope, level):
        """Constructor method."""
        self.group_id = group_id.strip()[0:ID_LEN]
        self.scope = scope
        self.level = level

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<GroupPerm('{group_id}', '{scope}', '{level}')>".format(
            group_id=self.group_id, scope=self.scope, level=self.level)



# $Id: __init__.py 9a2faf8ae708 2012/02/21 10:45:50 patrick $
"""Version Control System library."""

from abc import ABCMeta, abstractmethod
from os import mkdir, remove
from os.path import exists, join, normpath
from locale import resetlocale
from datetime import datetime
from urlparse import urlparse, urlunparse

from pyramid.i18n import TranslationString

from ...lib.utils import _, EXCLUDED_FILES


# =============================================================================
class Vcs(object):
    """Abstract base class for version control class."""

    __metaclass__ = ABCMeta
    engine = None

    # -------------------------------------------------------------------------
    def __init__(self, path, url, user_id=None, password=None):
        """Constructor method.

        :param path: (string)
             Path to local copy.
        :param url: (string)
            URL of source repository.
        :param user_id: (string, optional)
            User ID for clone/pull access.
        :param password: (string, optional)
            Password for clone/pull access.
        """
        resetlocale()
        self.path = str(normpath(path))
        self.url = str(url)
        self._user_id = str(user_id or '')
        self._password = str(password or '')

    # -------------------------------------------------------------------------
    def __repr__(self):
        """String representation."""
        return "<Vcs('%s', '%s', '%s', '%s')>" % (
            self.engine, self.path, self.url, self._user_id)

    # -------------------------------------------------------------------------
    @abstractmethod
    def clone(self, handler=None):
        """Create a copy of an existing repository in a directory. (abstract)

        :param handler: (:class:`~.lib.handler.Handler` instance, optional)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def pull_update(self, handler):
        """Pull changes and update. (abstract)

        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def commit_push(self, message, user_id, password, name, handler):
        """Commit and push changes. (abstract)

        :param message: (string)
            Commit message.
        :param user_id: (string)
            User ID for VCS access.
        :param password: (string)
            Cleared password for VCS access.
        :param name: (string)
            Name for VCS access.
        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def revert_all(self, handler):
        """Revert all files of the repository. (abstract)

        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def last_change(self):
        """Return the last change on the repository. (abstract)

        :return: (tuple)
           A tuple such as ``(date, changeset, user)``.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def file_infos(self, path, filename):
        """Return information on a file. (abstract)

        :param path: (string)
            Relative path.
        :param filename: (string)
            Name of the file to read.
        :return: (tuple)
            A tuple such as ``(date, changeset, user, message)``.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def log(self, path, filename, limit=1):
        """show revision history of file ``filename``.

        :param path: (string)
            Relative path to file.
        :param filename: (string)
            Name of the file.
        :param limit: (integer, default=1)
            Maximum number of entries in log.
        :return: (list of tuples or string)
            Log or error message.

        Each tuple or entry is like ``(date, changeset, user, message)``.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def add(self, path, handler):
        """Add all new files in ``path``. (abstract)

        :param path: (string)
            Relative path to browse.
        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    def mkdir(self, path, name, handler):
        """Make the directroy ``name``.

        :param path: (string)
            Relative path to directory to create.
        :param name: (string)
            Name of directory to create.
        """
        name = self.full_path(path, name)
        if isinstance(name, TranslationString):
            return handler.report('error', name)
        if not exists(name):
            mkdir(name)

    # -------------------------------------------------------------------------
    @abstractmethod
    def rename(self, path, filename, new_name, handler):
        """Rename a file. (abstract)

        :param path: (string)
            Relative path to file to rename.
        :param filename: (string)
            Name of the file to remove.
        :param new_new: (string)
            New name.
        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def remove(self, path, filename, handler):
        """Remove a file. (abstract)

        :param path: (string)
            Relative path to file to remove.
        :param filename: (string)
            Name of the file to remove.
        :param handler: (:class:`~.lib.handler.Handler` instance)
            Owner of this action.
        :return: (string)
            Error message or ``None`` if it succeeds.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def revision(self, fullname, revision):
        """Retrieve a revision. (abstract)

        :param fullname: (string)
            Full name of the file.
        :param revision: (string)
            Revision number to retrieve.
        :return: (string)
            Content of the file.
        """
        pass

    # -------------------------------------------------------------------------
    @abstractmethod
    def diff(self, fullname, revision):
        """Return differences between revision ``revision`` and current
        revision. (abstract)

        :param fullname: (string)
            Full name of the file.
        :param revision: (string)
            Revision number to compare.
        :return: (string)
            Differences.
        """
        pass

    # -------------------------------------------------------------------------
    def full_path(self, *path):
        """Return normalized full path of ``path`` file or an error message if
        it is outside the storage.

        :param path: (strings)
            Path chunks.
        :return: (string or :class:`pyramid.i18n.TranslationString` instance)
            Full path or error message.
        """
        full_path = normpath(join(self.path, *path))
        if not full_path.startswith(self.path):
            full_path = self.path
        return full_path.encode('utf8')

    # -------------------------------------------------------------------------
    def _log_path(self, path, filename):
        """Return normalized full path of file or ``None``.

        :param path: (string)
            Relative path to file.
        :param filename: (string)
            Name of file.
        :return: (string or ``None``)
            Normalized full path or ``None`` if the file is not eligible.
        """
        if filename in EXCLUDED_FILES:
            return
        filename = normpath(join(self.path, path, filename))
        if not filename.startswith(self.path) or not exists(filename):
            return
        return filename.encode('utf8')

    # -------------------------------------------------------------------------
    def _full_url(self, user_id=None, password=None):
        """Return an URL with user_id and password.

        :param user_id: (string)
            User ID for VCS access.
        :param password: (string)
            Password ID for VCS access.
        :return: (string)
            Full URL.
        """
        user_id = user_id or self._user_id
        if not user_id:
            return self.url
        scheme, netloc, path, params, query, fragment = urlparse(str(self.url))
        netloc = '%s:%s@%s' % (user_id, password or self._password,
                               netloc.rpartition('@')[2])
        return urlunparse((scheme, netloc, path, params, query, fragment))
